from __future__ import annotations

import contextlib
import os
from tempfile import TemporaryFile
from typing import List, Union

import numpy as np
from ase import Atoms
from ase.calculators.calculator import Calculator, all_changes
from ase.stress import full_3x3_to_voigt_6_stress

import _nepy
from calorine.nep.model import _get_nep_contents
from calorine.nep.nep import _check_components_polarizability_gradient, \
        _polarizability_gradient_to_3x3


class CPUNEP(Calculator):
    """This class provides an ASE calculator for `nep_cpu`,
    the in-memory CPU implementation of GPUMD.

    Parameters
    ----------
    model_filename : str
        Path to file in ``nep.txt`` format with model parameters
    atoms : Atoms
        Atoms to attach the calculator to
    label : str
        Label for this calclator
    debug : bool, optional
        Flag to toggle debug mode. Prints GPUMD output. Defaults to False.

    Raises
    ------
    FileNotFoundError
        Raises :class:`FileNotFoundError` if :attr:`model_filename` does not point to a valid file.
    ValueError
        Raises :class:`ValueError` atoms are not defined when trying to get energies and forces.
    Example
    -------

    >>> calc = CPUNEP('nep.txt')
    >>> atoms.calc = calc
    >>> atoms.get_potential_energy()
    """

    implemented_properties = [
        'energy',
        'energies',
        'forces',
        'stress',
    ]
    debug = False
    nepy = None

    def __init__(
        self,
        model_filename: str,
        atoms: Atoms | None = None,
        label: str | None = None,
        debug: bool = False,
    ):
        self.debug = debug

        if not os.path.exists(model_filename):
            raise FileNotFoundError(f'{model_filename} does not exist.')
        self.model_filename = str(model_filename)

        # Get model type from first row in nep.txt
        header, _ = _get_nep_contents(self.model_filename)
        self.model_type = header['model_type']
        self.supported_species = set(header['types'])
        self.nep_version = header['version']

        if self.model_type == 'dipole':
            # Only available for dipole models
            self.implemented_properties = ['dipole']
        elif self.model_type == 'polarizability':
            # Only available for polarizability models
            self.implemented_properties = ['polarizability']

        # Initialize atoms, results and nepy - note that this is also done in Calculator.__init__()
        if atoms is not None:
            self.set_atoms(atoms)
        parameters = {'model_filename': model_filename}
        Calculator.__init__(self, label=label, atoms=atoms, **parameters)
        if atoms is not None:
            self._setup_nepy()

    def __str__(self) -> str:
        def indent(s: str, i: int) -> str:
            s = '\n'.join([i * ' ' + line for line in s.split('\n')])
            return s

        parameters = '\n'.join(
            [f'{key}: {value}' for key, value in self.parameters.items()]
        )
        parameters = indent(parameters, 4)
        using_debug = '\nIn debug mode' if self.debug else ''

        s = f'{self.__class__.__name__}\n{parameters}{using_debug}'
        return s

    def _setup_nepy(self):
        """
        Creates an instance of the NEPY class and attaches it to the calculator object.
        The output from `nep.cpp` is only written to STDOUT if debug == True
        """
        if self.atoms is None:
            raise ValueError('Atoms must be defined to get energies and forces.')
        if self.atoms.cell.rank == 0:
            raise ValueError('Atoms must have a defined cell.')

        natoms = len(self.atoms)
        self.natoms = natoms
        c = self.atoms.get_cell(complete=True).flatten()
        cell = [c[0], c[3], c[6], c[1], c[4], c[7], c[2], c[5], c[8]]
        symbols = self.atoms.get_chemical_symbols()
        positions = list(
            self.atoms.get_positions().T.flatten()
        )  # [x1, ..., xN, y1, ... yN,...]
        masses = self.atoms.get_masses()

        # Disable output from C++ code by default
        if self.debug:
            self.nepy = _nepy.NEPY(
                self.model_filename, self.natoms, cell, symbols, positions, masses
            )
        else:
            with TemporaryFile('w') as f:
                with contextlib.redirect_stdout(f):
                    self.nepy = _nepy.NEPY(
                        self.model_filename,
                        self.natoms,
                        cell,
                        symbols,
                        positions,
                        masses,
                    )

    def set_atoms(self, atoms: Atoms):
        """Updates the Atoms object.

        Parameters
        ----------
        atoms : Atoms
            Atoms to attach the calculator to
        """
        species_in_atoms_object = set(np.unique(atoms.get_chemical_symbols()))
        if not species_in_atoms_object.issubset(self.supported_species):
            raise ValueError('Structure contains species that are not supported by the NEP model.')
        self.atoms = atoms
        self.results = {}
        self.nepy = None

    def _update_symbols(self):
        """Update atom symbols in NEPY."""
        symbols = self.atoms.get_chemical_symbols()
        self.nepy.set_symbols(symbols)

    def _update_masses(self):
        """Update atom masses in NEPY"""
        masses = self.atoms.get_masses()
        self.nepy.set_masses(masses)

    def _update_cell(self):
        """Update cell parameters in NEPY."""
        c = self.atoms.get_cell(complete=True).flatten()
        cell = [c[0], c[3], c[6], c[1], c[4], c[7], c[2], c[5], c[8]]
        self.nepy.set_cell(cell)

    def _update_positions(self):
        """Update atom positions in NEPY."""
        positions = list(
            self.atoms.get_positions().T.flatten()
        )  # [x1, ..., xN, y1, ... yN,...]
        self.nepy.set_positions(positions)

    def calculate(
        self,
        atoms: Atoms = None,
        properties: List[str] = None,
        system_changes: List[str] = all_changes,
    ):
        """Calculate energy, per atom energies, forces, stress and dipole.

        Parameters
        ----------
        atoms : Atoms, optional
            System for which to calculate properties, by default None
        properties : List[str], optional
            Properties to calculate, by default None
        system_changes : List[str], optional
            Changes to the system since last call, by default all_changes
        """
        if properties is None:
            properties = self.implemented_properties

        Calculator.calculate(self, atoms, properties, system_changes)

        if self.nepy is None:
            # Create new NEPY interface
            self._setup_nepy()
        # Update existing NEPY interface
        for change in system_changes:
            if change == 'positions':
                self._update_positions()
            elif change == 'numbers':
                self._update_symbols()
                self._update_masses()
            elif change == 'cell':
                self._update_cell()

        if 'dipole' in self.implemented_properties:
            dipole = np.array(self.nepy.get_dipole())
            self.results['dipole'] = dipole
        elif 'polarizability' in self.implemented_properties:
            pol = np.array(self.nepy.get_polarizability())
            polarizability = np.array([
                    [pol[0], pol[3], pol[5]],
                    [pol[3], pol[1], pol[4]],
                    [pol[5], pol[4], pol[2]]
                ])
            self.results['polarizability'] = polarizability
        else:
            energies, forces, virials = self.nepy.get_potential_forces_and_virials()
            energies_per_atom = np.array(energies)
            energy = energies_per_atom.sum()
            forces_per_atom = np.array(forces).reshape(-1, self.natoms).T
            virials_per_atom = np.array(virials).reshape(-1, self.natoms).T
            stress = -(np.sum(virials_per_atom, axis=0) / self.atoms.get_volume()).reshape((3, 3))
            stress = full_3x3_to_voigt_6_stress(stress)

            self.results['energy'] = energy
            self.results['forces'] = forces_per_atom
            self.results['stress'] = stress

    def get_dipole_gradient(
        self,
        displacement: float = 0.01,
        method: str = 'central difference',
        charge: float = 1.0,
    ):
        """Calculates the dipole gradient using finite differences.

        Parameters
        ----------
        displacement
            Displacement in Å to use for finite differences. Defaults to 0.01 Å.
        method
            Method for computing gradient with finite differences.
            One of 'forward difference' and 'central difference'.
            Defaults to 'central difference'
        charge
            System charge in units of the elemental charge.
            Used for correcting the dipoles before computing the gradient.
            Defaults to 1.0.

        Returns
        ------- dipole gradient with shape `(N, 3, 3)`
        """
        if 'dipole' not in self.implemented_properties:
            raise ValueError('Dipole gradients are only defined for dipole NEP models.')

        if displacement <= 0:
            raise ValueError('displacement must be > 0 Å')

        implemented_methods = {
            'forward difference': 0,
            'central difference': 1,
            'second order central difference': 2,
        }

        if method not in implemented_methods.keys():
            raise ValueError(f'Invalid method {method} for calculating gradient')

        if self.nepy is None:
            # Create new NEPY interface
            self._setup_nepy()

        dipole_gradient = np.array(
            self.nepy.get_dipole_gradient(
                displacement, implemented_methods[method], charge
            )
        ).reshape(self.natoms, 3, 3)
        return dipole_gradient

    def get_polarizability(
        self,
        atoms: Atoms = None,
        properties: List[str] = None,
        system_changes: List[str] = all_changes,
    ) -> np.ndarray:
        """Calculates the polarizability tensor for the current structure.
        The model must have been trained to predict the polarizability.
        This is a wrapper function for :func:`calculate`.

        Parameters
        ----------
        atoms : Atoms, optional
            System for which to calculate properties, by default None
        properties : List[str], optional
            Properties to calculate, by default None
        system_changes : List[str], optional
            Changes to the system since last call, by default all_changes

        Returns
        -------
        polarizability with shape ``(3, 3)``
        """
        if properties is None:
            properties = self.implemented_properties

        if 'polarizability' not in properties:
            raise ValueError('Polarizability is only defined for polarizability NEP models.')
        self.calculate(atoms, properties, system_changes)
        return self.results['polarizability']

    def get_polarizability_gradient(
        self,
        displacement: float = 0.01,
        component: Union[str, List[str]] = 'full',
    ) -> np.ndarray:
        """Calculates the dipole gradient for a given structure using finite differences.
        This function computes the derivatives using the second-order central difference
        method with a C++ backend.

        Parameters
        ----------
        displacement
            Displacement in Å to use for finite differences. Defaults to ``0.01``.
        component
            Component or components of the polarizability tensor that the gradient
            should be computed for. The following components are available:
                ``x`, ``y``, ``z``, ``full``
            Option ``full`` computes the derivative whilst moving the atoms in each Cartesian
            direction, which yields a tensor of shape (N, 3, 3, 3).
            Multiple components may be specified.
            Defaults to ``full``.


        Returns
        -------
            polarizability gradient with shape ``(N, C, 3, 3)`` where ``C``
            are the number of components chosen.
        """
        if 'polarizability' not in self.implemented_properties:
            raise ValueError('Polarizability gradients are only defined'
                             ' for polarizability NEP models.')

        if displacement <= 0:
            raise ValueError('displacement must be > 0 Å')

        if self.nepy is None:
            # Create new NEPY interface
            self._setup_nepy()

        component_array = _check_components_polarizability_gradient(component)

        pg = np.array(
            self.nepy.get_polarizability_gradient(
                displacement, component_array
            )
        ).reshape(self.natoms, 3, 6)
        polarizability_gradient = _polarizability_gradient_to_3x3(self.natoms, pg)
        return polarizability_gradient[:, component_array, :, :]



from os.path import exists
from os.path import join as join_path
from typing import Any, Dict, Iterable, List, NamedTuple, TextIO, Tuple
from warnings import warn

import numpy as np
from ase import Atoms
from ase.io import read, write
from ase.stress import voigt_6_to_full_3x3_stress
from ase.units import GPa
from pandas import DataFrame


def read_loss(filename: str) -> DataFrame:
    """Parses a file in ``loss.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found `here
    <https://gpumd.org/nep/output_files/loss_out.html>`__.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in loss.out, append a dimension
        data = data.reshape(1, -1)
    if len(data[0]) == 6:
        tags = 'total_loss L1 L2'
        tags += ' RMSE_P_train'
        tags += ' RMSE_P_test'
    elif len(data[0]) == 10:
        tags = 'total_loss L1 L2'
        tags += ' RMSE_E_train RMSE_F_train RMSE_V_train'
        tags += ' RMSE_E_test RMSE_F_test RMSE_V_test'
    else:
        raise ValueError(
            f'Input file contains {len(data[0])} data columns. Expected 6 or 10 columns.'
        )
    generations = range(100, len(data) * 100 + 1, 100)
    df = DataFrame(data=data[:, 1:], columns=tags.split(), index=generations)
    return df


def _write_structure_in_nep_format(structure: Atoms, f: TextIO) -> None:
    """Write structure block into a file-like object in format readable by nep executable.

    Parameters
    ----------
    structure
        input structure; must hold information regarding energy and forces
    f
        file-like object to which to write
    """

    # Allowed keyword=value pairs. Use ASEs extyz write functionality.:
    #   lattice="ax ay az bx by bz cx cy cz"                    (mandatory)
    #   energy=energy_value                                     (mandatory)
    #   virial="vxx vxy vxz vyx vyy vyz vzx vzy vzz"            (optional)
    #   weight=relative_weight                                  (optional)
    #   properties=property_name:data_type:number_of_columns
    #       species:S:1                      (mandatory)
    #       pos:R:3                          (mandatory)
    #       force:R:3 or forces:R:3          (mandatory)
    try:
        structure.get_potential_energy()
        structure.get_forces()  # calculate forces to have them on the Atoms object
    except RuntimeError:
        raise RuntimeError('Failed to retrieve energy and/or forces for structure')
    if np.isclose(structure.get_volume(), 0):
        raise ValueError('Structure cell must have a non-zero volume!')
    try:
        structure.get_stress()
    except RuntimeError:
        warn('Failed to retrieve stresses for structure')
    write(filename=f, images=structure, write_info=True, format='extxyz')


def write_structures(outfile: str, structures: List[Atoms]) -> None:
    """Writes structures for training/testing in format readable by nep executable.

    Parameters
    ----------
    outfile
        output filename
    structures
        list of structures with energy, forces, and (possibly) stresses
    """
    with open(outfile, 'w') as f:
        for structure in structures:
            _write_structure_in_nep_format(structure, f)


def write_nepfile(parameters: NamedTuple, dirname: str) -> None:
    """Writes parameters file for NEP construction.

    Parameters
    ----------
    parameters
        input parameters; see `here <https://gpumd.org/nep/input_parameters/index.html>`__
    dirname
        directory in which to place input file and links
    """
    with open(join_path(dirname, 'nep.in'), 'w') as f:
        for key, val in parameters.items():
            f.write(f'{key}  ')
            if isinstance(val, Iterable):
                f.write(' '.join([f'{v}' for v in val]))
            else:
                f.write(f'{val}')
            f.write('\n')


def read_nepfile(filename: str) -> Dict[str, Any]:
    """Returns the content of a configuration file (`nep.in`) as a dictionary.

    Parameters
    ----------
    filename
        input file name
    """
    settings = {}
    with open(filename) as f:
        for line in f.readlines():
            flds = line.split()
            if len(flds) == 0:
                continue
            if flds[0].startswith('#'):
                continue
            settings[flds[0]] = ' '.join(flds[1:])
    for key, val in settings.items():
        if key in ['version', 'neuron', 'generation', 'batch', 'population', 'mode', 'model_type']:
            settings[key] = int(val)
        elif key in [
            'lambda_1',
            'lambda_2',
            'lambda_e',
            'lambda_f',
            'lambda_v',
            'lambda_shear',
            'force_delta',
        ]:
            settings[key] = float(val)
        elif key in ['cutoff', 'n_max', 'l_max', 'basis_size', 'zbl', 'type_weight']:
            settings[key] = [float(v) for v in val.split()]
        elif key == 'type':
            types = val.split()
            types[0] = int(types[0])
            settings[key] = types
    return settings


def read_structures(dirname: str) -> Tuple[List[Atoms], List[Atoms]]:
    """Parses the ``energy_*.out``, ``force_*.out``, ``virial_*.out``,
    ``polarizability_*.out`` and ``dipole_*.out`` files from a nep run and
    returns their content as lists. The first and second list contain the structures
    from the training and test sets, respectively. Each list entry corresponds to an ASE
    Atoms object, which in turn contains predicted and target energies, forces and virials/stresses
    or polarizability/diople stored in the `info` property.

    Parameters
    ----------
    dirname
        directory from which to read output files

    """
    path = join_path(dirname)
    if not exists(path):
        raise ValueError(f'Directory {path} does not exist')
    nep_info = read_nepfile(f'{path}/nep.in')
    if 'mode' in nep_info or 'model_type' in nep_info:
        ftype = nep_info.get('mode', nep_info.get('model_type'))
        if ftype == 2 or ftype == 1:
            return _read_structures_tensors(dirname, ftype)
    return _read_structures_potential(dirname)


def _read_structures_tensors(dirname: str, ftype: int) \
        -> Tuple[List[Atoms], List[Atoms]]:
    """Parses the ``polarizability_*.out`` and ``dipole_*.out``
    files from a nep run and returns their content as lists.
    The first and second list contain the structures from the training and
    test sets, respectively. Each list entry corresponds to an ASE
    Atoms object, which in turn contains predicted and target
    dipole or polarizability stored in the `info`
    property.

    Parameters
    ----------
    dirname
        directory from which to read output files

    """
    path = join_path(dirname)
    structures = {}

    if ftype == 1:
        sname = 'dipole'
    else:
        sname = 'polarizability'

    for stype in ['train', 'test']:
        filename = join_path(dirname, f'{stype}.xyz')
        try:
            structures[stype] = read(filename, format='extxyz', index=':')
        except FileNotFoundError:
            warn(f'File {filename} not found.')
            structures[stype] = []
            continue

        n_structures = len(structures[stype])
        ts, ps = _read_data_file(path, f'{sname}_{stype}.out')
        if ftype == 1:
            ts = np.array(ts).reshape((-1, 3))
            ps = np.array(ps).reshape((-1, 3))
        else:
            ts = np.array(ts).reshape((-1, 6))
            ps = np.array(ps).reshape((-1, 6))
        assert len(ts) == n_structures, \
            f'Number of structures in {sname}_{stype}.out ({len(ts)})' \
            f' and {stype}.xyz ({n_structures}) inconsistent'
        for structure, t, p in zip(structures[stype], ts, ps):
            if ftype == 1:
                assert np.shape(t) == (3,)
                assert np.shape(p) == (3,)
            else:
                assert np.shape(t) == (6,)
                assert np.shape(p) == (6,)
            structure.info[f'{sname}_target'] = t
            structure.info[f'{sname}_predicted'] = p

    return structures['train'], structures['test']


def _read_structures_potential(dirname: str) -> Tuple[List[Atoms], List[Atoms]]:
    """Parses the ``energy_*.out``, ``force_*.out``, ``virial_*.out``
    files from a nep run and returns their content as lists.
    The first and second list contain the structures from the training and
    test sets, respectively. Each list entry corresponds to an ASE
    Atoms object, which in turn contains predicted and target
    energies, forces and virials/stresses stored in the `info`
    property.

    Parameters
    ----------
    dirname
        directory from which to read output files

    """
    path = join_path(dirname)
    structures = {}

    for stype in ['train', 'test']:
        file_path = join_path(dirname, f'{stype}.xyz')
        if not exists(file_path):
            warn(f'File {file_path} not found.')
            structures[stype] = []
            continue

        structures[stype] = read(
            file_path, format='extxyz', index=':'
        )

        ts, ps = _read_data_file(path, f'energy_{stype}.out')
        n_structures = len(structures[stype])
        assert len(ts) == n_structures, (
            f'Number of structures in energy_{stype}.out ({len(ts)})'
            f' and {stype}.xyz ({n_structures}) inconsistent'
        )
        for structure, t, p in zip(structures[stype], ts, ps):
            structure.info['energy_target'] = t
            structure.info['energy_predicted'] = p

        ts, ps = _read_data_file(path, f'force_{stype}.out')
        n_atoms_total = sum([len(s) for s in structures[stype]])
        assert len(ts) == n_atoms_total, (
            f'Number of structures in force_{stype}.out ({len(ts)})'
            f' and {stype}.xyz ({n_structures}) inconsistent'
        )
        n = 0
        for structure in structures[stype]:
            nat = len(structure)
            structure.info['force_target'] = np.array(ts[n: n + nat]).reshape(nat, 3)
            structure.info['force_predicted'] = np.array(ps[n: n + nat]).reshape(
                nat, 3
            )
            n += nat

        ts, ps = _read_data_file(path, f'virial_{stype}.out')
        ts, ps = np.array(ts), np.array(ps)
        N = len(structures[stype])
        if ts.shape == (6*N,):
            # GPUMD <=v3.6 style virial_*.out
            # First column are NEP predictions, second are targets
            # Order: First N values are xx, second are yy etc.
            ts = np.array(ts).reshape((6, -1)).T  # GPUMD 3.6 compatibility
            ps = np.array(ps).reshape((6, -1)).T
        elif ts.shape == (N, 6):
            # GPUMD >=v3.7 style virial_*.out
            # First 6 columns are NEP predictions, last 6 are targets
            # Order: xx, yy, zz, xy, yz, zx
            pass
        else:
            raise ValueError(f'virial_*.out has invalid shape, {ts.shape}')

        assert len(ts) == n_structures, \
            f'Number of structures in virial_{stype}.out ({len(ts)})' \
            f' and {stype}.xyz ({n_structures}) inconsistent'
        for structure, t, p in zip(structures[stype], ts, ps):
            assert np.shape(t) == (6,)
            structure.info['virial_target'] = t
            structure.info['virial_predicted'] = p
            conv = len(structure) / structure.get_volume() / GPa
            structure.info['stress_target'] = t * conv
            structure.info['stress_predicted'] = p * conv

    return structures['train'], structures['test']


def _read_data_file(dirname: str, fname: str):
    """Private function that parses energy/force/virial_*.out files and
    returns their content for further processing.
    """
    path = join_path(dirname, fname)
    if not exists(path):
        raise ValueError(f'Directory {path} does not exist')
    with open(path, 'r') as f:
        lines = f.readlines()
    target, predicted = [], []
    for line in lines:
        flds = line.split()
        if len(flds) == 12:  # Virial after GPUMD 3.7
            predicted.append([float(s) for s in flds[0:6]])
            target.append([float(s) for s in flds[6:12]])
        elif len(flds) == 6:  # Force
            predicted.append([float(s) for s in flds[0:3]])
            target.append([float(s) for s in flds[3:6]])
        elif len(flds) == 2:  # Energy, virial before GPUMD 3.7
            predicted.append(float(flds[0]))
            target.append(float(flds[1]))
        else:
            raise ValueError(f'Malformed file: {path}')
    return target, predicted


def get_parity_data(
    structures: List[Atoms],
    property: str,
    selection: List[str] = None,
    flatten: bool = True,
) -> DataFrame:
    """Returns the predicted and target energies, forces, virials or stresses
    from a list of structures in a format suitable for generating parity plots.

    The structures should  have been read using :func:`read_structures
    <calorine.nep.read_structures>`, such that the ``info``-object is
    populated with keys on the form ``<property>_<type>`` where ``<property>``
    is one of ``energy``, ``force``, ``virial``, and stress, and ``<type>`` is one
    of ``predicted`` or ``target``.

    The resulting parity data is returned as a tuple of dicts, where each entry
    corresponds to a list.

    Parameters
    ----------
    structures
        List of structures as read with :func:`read_structures <calorine.nep.read_structures>`.
    property
        One of ``energy``, ``force``, ``virial``, ``stress``, ``polarizability``, ``dipole``.
    selection
        A list containing which components to return, and/or the absolute value.
        Possible values are ``x``, ``y``, ``z``, ``xx``, ``yy``,
        ``zz``, ``yz``, ``xz``, ``xy``, ``abs``, ``pressure``.
    flatten
        if True return flattened lists; this is useful for flattening
        the components of force or virials into a simple list
    """
    data = {'predicted': [], 'target': []}
    voigt_mapping = {
        'x': 0,
        'y': 1,
        'z': 2,
        'xx': 0,
        'yy': 1,
        'zz': 2,
        'yz': 3,
        'xz': 4,
        'xy': 5,
    }
    if property not in ('energy', 'force', 'virial', 'stress', 'polarizability', 'dipole'):
        raise ValueError(
            "`property` must be one of 'energy', 'force', 'virial', 'stress',"
            " 'polarizability', 'dipole'."
        )
    if property == 'energy' and selection:
        raise ValueError('Selection does nothing for scalar-valued `energy`.')
    if property != 'stress' and selection and 'pressure' in selection:
        raise ValueError(f'Cannot calculate pressure for `{property}`.')
    for structure in structures:
        for stype in data:
            property_with_stype = f'{property}_{stype}'
            if property_with_stype not in structure.info.keys():
                raise KeyError(f'{property_with_stype} does not exist in info object!')
            extracted_property = np.array(structure.info[property_with_stype])

            if selection is None or len(selection) == 0:
                data[stype].append(extracted_property)
                continue

            selected_values = []
            for select in selection:
                if property == 'force':
                    # flip to get (n_components, n_structures)
                    extracted_property = extracted_property.T
                if select == 'abs':
                    if property == 'force':
                        selected_values.append(np.linalg.norm(extracted_property, axis=0))
                    else:
                        # property can only be in ('virial', 'stress')
                        full_tensor = voigt_6_to_full_3x3_stress(extracted_property)
                        selected_values.append(np.linalg.norm(full_tensor))
                    continue

                if select == 'pressure' and property == 'stress':
                    total_stress = extracted_property
                    selected_values.append(-np.sum(total_stress[:3]) / 3)
                    continue

                if select not in voigt_mapping:
                    raise ValueError(f'Selection `{select}` is not allowed.')
                index = voigt_mapping[select]
                if index >= extracted_property.shape[0]:
                    raise ValueError(
                        f'Selection `{select}` is not compatible with property `{property}`.'
                    )
                selected_values.append(extracted_property[index])
            data[stype].append(selected_values)
    if flatten:
        for key, value in data.items():
            if len(np.shape(value[0])) > 0:
                data[key] = np.concatenate(value).ravel().tolist()
    df = DataFrame(data)
    # In case of flatten, cast to float64 for compatibility
    # with e.g. seaborn.
    # Casting in this way breaks tensorial properties though,
    # so skip it there.
    if flatten:
        df['target'] = df.target.astype('float64')
        df['predicted'] = df.predicted.astype('float64')
    return df



from warnings import warn
from collections.abc import Iterable
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
from ase import Atoms
from ase.io import read, write
from pandas import DataFrame


def read_kappa(filename: str) -> DataFrame:
    """Parses a file in ``kappa.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found `here
    <https://gpumd.org/gpumd/output_files/kappa_out.html>`__.

    Parameters
    ----------
    filename
        Input file name.
    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in kappa.out, append a dimension
        data = data.reshape(1, -1)
    tags = 'kx_in kx_out ky_in ky_out kz_tot'.split()
    if len(data[0]) != len(tags):
        raise ValueError(
            f'Input file contains {len(data[0])} data columns.'
            f' Expected {len(tags)} columns.'
        )
    df = DataFrame(data=data, columns=tags)
    df['kx_tot'] = df.kx_in + df.kx_out
    df['ky_tot'] = df.ky_in + df.ky_out
    return df


def read_hac(filename: str) -> DataFrame:
    """Parses a file in ``hac.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found `here
    <https://gpumd.org/gpumd/output_files/hac_out.html>`__.

    Parameters
    ----------
    filename
        Input file name.
    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in hac.out, append a dimension
        data = data.reshape(1, -1)
    tags = 'time'
    tags += ' jin_jtot_x jout_jtot_x jin_jtot_y jout_jtot_y jtot_jtot_z'
    tags += ' kx_in kx_out ky_in ky_out kz_tot'
    tags = tags.split()
    if len(data[0]) != len(tags):
        raise ValueError(
            f'Input file contains {len(data[0])} data columns.'
            f' Expected {len(tags)} columns.'
        )
    df = DataFrame(data=data, columns=tags)
    df['kx_tot'] = df.kx_in + df.kx_out
    df['ky_tot'] = df.ky_in + df.ky_out
    # remove columns with less relevant data to save space
    for col in df:
        if 'jtot' in col or '_in' in col:
            del df[col]
    return df


def read_thermo(filename: str, natoms: int = 1) -> DataFrame:
    """Parses a file in ``thermo.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found `here
    <https://gpumd.org/gpumd/output_files/thermo_out.html>`__.

    Parameters
    ----------
    filename
        Input file name.
    natoms
        Number of atoms; used to normalize energies.
    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in loss.out, append a dimension
        data = data.reshape(1, -1)
    if len(data[0]) == 9:
        # orthorhombic box
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz'
        tags += ' cell_xx cell_yy cell_zz'
    elif len(data[0]) == 12:
        # orthorhombic box with stresses in Voigt notation (v3.3.1+)
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz stress_yz stress_xz stress_xy'
        tags += ' cell_xx cell_yy cell_zz'
    elif len(data[0]) == 15:
        # triclinic box
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz'
        tags += (
            ' cell_xx cell_xy cell_xz cell_yx cell_yy cell_yz cell_zx cell_zy cell_zz'
        )
    elif len(data[0]) == 18:
        # triclinic box with stresses in Voigt notation (v3.3.1+)
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz stress_yz stress_xz stress_xy'
        tags += (
            ' cell_xx cell_xy cell_xz cell_yx cell_yy cell_yz cell_zx cell_zy cell_zz'
        )
    else:
        raise ValueError(
            f'Input file contains {len(data[0])} data columns.'
            ' Expected 9, 12, 15 or 18 columns.'
        )
    df = DataFrame(data=data, columns=tags.split())
    assert natoms > 0, 'natoms must be positive'
    df.kinetic_energy /= natoms
    df.potential_energy /= natoms
    return df


def read_xyz(filename: str) -> Atoms:
    """
    Reads the structure input file (``model.xyz``) for GPUMD and returns the
    structure.

    This is a wrapper function around :func:`ase.io.read_xyz` since the ASE implementation does
    not read velocities properly.

    Parameters
    ----------
    filename
        Name of file from which to read the structure.

    Returns
    -------
        Structure as ASE Atoms object with additional per-atom arrays
        representing atomic masses, velocities etc.
    """
    structure = read(filename, format='extxyz')
    if structure.has('vel'):
        structure.set_velocities(structure.get_array('vel'))
    return structure


def read_runfile(filename: str) -> List[Tuple[str, list]]:
    """
    Parses a GPUMD input file in ``run.in`` format and returns the
    content in the form a list of keyword-value pairs.

    Parameters
    ----------
    filename
        Input file name.

    Returns
    -------
        List of keyword-value pairs.
    """
    data = []
    with open(filename, 'r') as f:
        for k, line in enumerate(f.readlines()):
            flds = line.split()
            if len(flds) == 0:
                continue
            elif len(flds) == 1:
                raise ValueError(f'Line {k} contains only one field:\n{line}')
            keyword = flds[0]
            values = tuple(flds[1:])
            if keyword in ['time_step', 'velocity']:
                values = float(values[0])
            elif keyword in ['dump_thermo', 'dump_position', 'dump_restart', 'run']:
                values = int(values[0])
            elif len(values) == 1:
                values = values[0]
            data.append((keyword, values))
    return data


def write_runfile(
    file: Path, parameters: List[Tuple[str, Union[int, float, Tuple[str, float]]]]
):
    """Write a file in run.in format to define input parameters for MD simulation.

    Parameters
    ----------
    file
        Path to file to be written.

    parameters : dict
        Defines all key-value pairs used in run.in file
        (see GPUMD documentation for a complete list).
        Values can be either floats, integers, or lists/tuples.
    """

    with open(file, 'w') as f:
        # Write all keywords with parameter(s)
        for key, val in parameters:
            f.write(f'{key} ')
            if isinstance(val, Iterable) and not isinstance(val, str):
                for v in val:
                    f.write(f'{v} ')
            else:
                f.write(f'{val}')
            f.write('\n')


def write_xyz(filename: str, structure: Atoms, groupings: List[List[List[int]]] = None):
    """
    Writes a structure into GPUMD input format (`model.xyz`).

    Parameters
    ----------
    filename
        Name of file to which the structure should be written.
    structure
        Input structure.
    groupings
        Groups into which the individual atoms should be divided in the form of
        a list of list of lists. Specifically, the outer list corresponds to
        the grouping methods, of which there can be three at the most, which
        contains a list of groups in the form of lists of site indices. The
        sum of the lengths of the latter must be the same as the total number
        of atoms.

    Raises
    ------
    ValueError
        Raised if parameters are incompatible.
    """
    # Make a local copy of the atoms object
    _structure = structure.copy()

    # Check velocties parameter
    velocities = _structure.get_velocities()
    if velocities is None or np.max(np.abs(velocities)) < 1e-6:
        has_velocity = 0
    else:
        has_velocity = 1

    # Check groupings parameter
    if groupings is None:
        number_of_grouping_methods = 0
    else:
        number_of_grouping_methods = len(groupings)
        if number_of_grouping_methods > 3:
            raise ValueError('There can be no more than 3 grouping methods!')
        for g, grouping in enumerate(groupings):
            all_indices = [i for group in grouping for i in group]
            if len(all_indices) != len(_structure) or set(all_indices) != set(
                range(len(_structure))
            ):
                raise ValueError(
                    f'The indices listed in grouping method {g} are'
                    ' not compatible with the input structure!'
                )

    # Allowed keyword=value pairs. Use ASEs extyz write functionality.
    #   pbc="pbc_a pbc_b pbc_c"
    #   lattice="ax ay az bx by bz cx cy cz"
    #   properties=property_name:data_type:number_of_columns
    #       species:S:1
    #       pos:R:3
    #       mass:R:1
    #       vel:R:3
    #       group:I:number_of_grouping_methods
    if _structure.has('mass'):
        # If structure already has masses set, use those
        warn('Structure already has array "mass"; will use existing values.')
    else:
        _structure.new_array('mass', _structure.get_masses())

    if has_velocity:
        _structure.new_array('vel', _structure.get_velocities())
    if groupings is not None:
        group_indices = np.array(
            [
                [
                    [
                        group_index
                        for group_index, group in enumerate(grouping)
                        if structure_idx in group
                    ]
                    for grouping in groupings
                ]
                for structure_idx in range(len(_structure))
            ]
        ).squeeze()  # pythoniccc
        _structure.new_array('group', group_indices)

    write(filename=filename, images=_structure, write_info=True, format='extxyz')


def read_mcmd(filename: str, accumulate: bool = True) -> DataFrame:
    """Parses a Monte Carlo output file in ``mcmd.out`` format
    and returns the content in the form of a DataFrame.

    Parameters
    ----------
    filename
        Path to file to be parsed.
    accumulate
        If ``True`` the MD steps between subsequent Monte Carlo
        runs in the same output file will be accumulated.

    Returns
    -------
        DataFrame containing acceptance ratios and concentrations (if available),
        as well as key Monte Carlo parameters.
    """
    with open(filename, 'r') as f:
        lines = f.readlines()
    data = []
    offset = 0
    step = 0
    accummulated_step = 0
    for line in lines:
        if line.startswith('# mc'):
            flds = line.split()
            mc_type = flds[2]
            md_steps = int(flds[3])
            mc_trials = int(flds[4])
            temperature_initial = float(flds[5])
            temperature_final = float(flds[6])
            if mc_type.endswith('sgc'):
                ntypes = int(flds[7])
                species = [flds[8+2*k] for k in range(ntypes)]
                phis = {f'phi_{flds[8+2*k]}': float(flds[9+2*k]) for k in range(ntypes)}
            kappa = float(flds[8+2*ntypes]) if mc_type == 'vcsgc' else np.nan
        elif line.startswith('# num_MD_steps'):
            continue
        else:
            flds = line.split()
            previous_step = step
            step = int(flds[0])
            if step <= previous_step and accumulate:
                offset += previous_step
            accummulated_step = step + offset
            record = dict(
                step=accummulated_step,
                mc_type=mc_type,
                md_steps=md_steps,
                mc_trials=mc_trials,
                temperature_initial=temperature_initial,
                temperature_final=temperature_final,
                acceptance_ratio=float(flds[1]),
            )
            if mc_type.endswith('sgc'):
                record.update(phis)
                if mc_type == 'vcsgc':
                    record['kappa'] = kappa
                concentrations = {f'conc_{s}': float(flds[k])
                                  for k, s in enumerate(species, start=2)}
                record.update(concentrations)
            data.append(record)
    df = DataFrame.from_dict(data)
    return df


def read_thermodynamic_data(directory_name: str) -> DataFrame:
    """Parses the data in a GPUMD output directory
    and returns the content in the form of a :class:`DataFrame`.
    This function reads the ``thermo.out`` and ``run.in`` files,
    and returns the thermodynamic data including the time (in ps),
    the pressure (in GPa), the side lengths of the simulation cell
    (in Å), and the volume (in Å:sup:`3`).

    Parameters
    ----------
    directory_name
        Path to directory to be parsed.

    Returns
    -------
        :class:`DataFrame` containing (augmented) thermodynamic data.
    """

    try:
        params = read_runfile(f'{directory_name}/run.in')
    except FileNotFoundError:
        raise FileNotFoundError(f'No `run.in` file found in {directory_name}')

    time_step, dump_thermo = None, None
    for p, v in params:
        if p == 'time_step':
            if time_step is not None:
                raise ValueError('There should only be a single occurrence'
                                 ' of the `time_step` keyword in `run.in`.')
            time_step = v
        elif p == 'dump_thermo':
            if dump_thermo is not None:
                raise ValueError('This function can only handle runs with a single'
                                 ' occurrence of the `dump_thermo` keyword in `dump.in`.')
            dump_thermo = v
    if time_step is None:
        raise ValueError('Could not extract value of `time_step` keyword from `run.in` file.')
    if dump_thermo is None:
        raise ValueError('Could not extract value of `dump_thermo` keyword from `run.in` file.')

    try:
        df = read_thermo(f'{directory_name}/thermo.out')
    except FileNotFoundError:
        raise FileNotFoundError(f'No `thermo.out` file found in {directory_name}')
    else:
        df['time'] = df.index * dump_thermo * time_step * 1e-3   # in ps
        df['pressure'] = (df.stress_xx + df.stress_yy + df.stress_zz) / 3
        if 'cell_xy' in df:
            df['alat'] = np.sqrt(df.cell_xx ** 2 + df.cell_xy ** 2 + df.cell_xz ** 2)
            df['blat'] = np.sqrt(df.cell_yx ** 2 + df.cell_yy ** 2 + df.cell_yz ** 2)
            df['clat'] = np.sqrt(df.cell_zx ** 2 + df.cell_zy ** 2 + df.cell_zz ** 2)
            volume = (df.cell_xx * df.cell_yy * df.cell_zz +
                      df.cell_xy * df.cell_yz * df.cell_xz +
                      df.cell_xz * df.cell_yx * df.cell_zy -
                      df.cell_xx * df.cell_yz * df.cell_zy -
                      df.cell_xy * df.cell_yx * df.cell_zz -
                      df.cell_xz * df.cell_yy * df.cell_zx)
        else:
            df['alat'] = df.cell_xx
            df['blat'] = df.cell_yy
            df['clat'] = df.cell_zz
            volume = (df.cell_xx * df.cell_yy * df.cell_zz)
        df['volume'] = volume

    return df



import contextlib
import os
import warnings
from pathlib import Path
from typing import List, Optional, Tuple, Union

import numpy as np
from ase import Atoms

import _nepy
from calorine.nep.model import _get_nep_contents


def _get_atomic_properties(
    structure: Atoms,
) -> Tuple[List[float], List[str], List[float]]:
    """Fetches cell, symbols and positions for a structure. Since NEP_CPU requires a cell, if the
    structure has no cell a default cubic cell with a side of 100 Å will be used.

    Parameters
    ----------
    structure
        Atoms object representing the structure.

    Returns
    -------
    List[float]
        Cell vectors
    List[str]
        Atomic species
    List[float]
        Atom positions
    List[float]
        Atom masses
    """
    if structure.cell.rank == 0:
        warnings.warn('Using default unit cell (cubic with side 100 Å).')
        set_default_cell(structure)

    c = structure.get_cell(complete=True).flatten()
    cell = [c[0], c[3], c[6], c[1], c[4], c[7], c[2], c[5], c[8]]

    symbols = structure.get_chemical_symbols()
    positions = list(
        structure.get_positions().T.flatten()
    )  # [x1, ..., xN, y1, ... yN,...]
    masses = structure.get_masses()
    return cell, symbols, positions, masses


def _setup_nepy(
    model_filename: str,
    natoms: int,
    cell: List[float],
    symbols: List[str],
    positions: List[float],
    masses: List[float],
    debug: bool,
) -> _nepy.NEPY:
    """Sets up an instance of the NEPY pybind11 interface to NEP_CPU.

    Parameters
    ----------
    model_filename
        Path to model.
    natoms:
        Number of atoms in the structure.
    cell:
        Cell vectors.
    symbols:
        Atom species.
    positions:
        Atom positions.
    masses:
        Atom masses.
    debug:
        Flag to control if the output from NEP_CPU will be printed.

    Returns
    -------
    NEPY
        NEPY interface
    """
    # Ensure that `model_filename` exists to avoid segfault in pybind11 code
    if not os.path.isfile(model_filename):
        raise ValueError(f'{Path(model_filename)} does not exist')

    # Disable output from C++ code by default
    if debug:
        nepy = _nepy.NEPY(model_filename, natoms, cell, symbols, positions, masses)
    else:
        with open(os.devnull, 'w') as f:
            with contextlib.redirect_stdout(f):
                with contextlib.redirect_stderr(f):
                    nepy = _nepy.NEPY(
                        model_filename, natoms, cell, symbols, positions, masses
                    )
    return nepy


def set_default_cell(structure: Atoms, box_length: float = 100):
    """Adds a cubic box to an Atoms object. Atoms object is edited in-place.

    Parameters
    ----------
    structure
        Structure to add box to
    box_length
        Cubic box side length in Å, by default 100
    """
    structure.set_cell([[box_length, 0, 0], [0, box_length, 0], [0, 0, box_length]])
    structure.center()


def get_descriptors(
    structure: Atoms, model_filename: Optional[str] = None, debug: bool = False
) -> np.ndarray:
    """Calculates the NEP descriptors for a given structure. A NEP model defined by a nep.txt
    can additionally be provided to get the NEP3 model specific descriptors.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model in ``nep.txt`` format. Defaults to ``None``.
    debug
        Flag to toggle debug mode. Makes the generated dummy NEP2 model available
        in a local tmp directory, as well as prints GPUMD output. Defaults to ``False``.

    Returns
    -------
    Descriptors for the supplied structure, with shape (natoms, descriptor components)
    """
    local_structure = structure.copy()
    natoms = len(local_structure)
    cell, symbols, positions, masses = _get_atomic_properties(local_structure)

    nepy = _setup_nepy(
        model_filename, natoms, cell, symbols, positions, masses, debug
    )
    all_descriptors = nepy.get_descriptors()
    descriptors_per_atom = np.array(all_descriptors).reshape(-1, natoms).T
    return descriptors_per_atom


def get_latent_space(
    structure: Atoms, model_filename: Union[str, None] = None, debug: bool = False
) -> np.ndarray:
    """Calculates the latent space representation of a structure, i.e, the activiations in
    the hidden layer. A NEP model defined by a `nep.txt` file needs to be provided.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model. Defaults to None.
    debug
        Flag to toggle debug mode. Prints GPUMD output. Defaults to False.

    Returns
    -------
    Activation with shape `(natoms, N_neurons)`
    """
    if model_filename is None:
        raise ValueError('Model is undefined')
    local_structure = structure.copy()
    natoms = len(local_structure)
    cell, symbols, positions, masses = _get_atomic_properties(local_structure)

    nepy = _setup_nepy(
        model_filename, natoms, cell, symbols, positions, masses, debug
    )

    latent = nepy.get_latent_space()
    latent = np.array(latent).reshape(-1, natoms).T
    return latent


def get_potential_forces_and_virials(
    structure: Atoms, model_filename: Optional[str] = None, debug: bool = False
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculates the per-atom potential, forces and virials for a given structure.
    A NEP model defined by a `nep.txt` file needs to be provided.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model. Defaults to None.
    debug
        Flag to toggle debug mode. Prints GPUMD output. Defaults to False.

    Returns
    -------
    potential with shape `(natoms,)`
    forces with shape `(natoms, 3)`
    virials with shape `(natoms, 9)`
    """
    if model_filename is None:
        raise ValueError('Model is undefined')

    model_type = _get_nep_contents(model_filename)[0]['model_type']
    if model_type != 'potential':
        raise ValueError(
            'A NEP model trained for predicting energies and forces must be used.'
        )

    local_structure = structure.copy()
    natoms = len(local_structure)
    cell, symbols, positions, masses = _get_atomic_properties(local_structure)

    nepy = _setup_nepy(
        model_filename, natoms, cell, symbols, positions, masses, debug
    )

    energies, forces, virials = nepy.get_potential_forces_and_virials()
    forces_per_atom = np.array(forces).reshape(-1, natoms).T
    virials_per_atom = np.array(virials).reshape(-1, natoms).T
    return np.array(energies), forces_per_atom, virials_per_atom


def get_polarizability(
    structure: Atoms,
    model_filename: Optional[str] = None,
    debug: bool = False,
) -> np.ndarray:
    """Calculates the polarizability tensor for a given structure. A NEP model defined
    by a ``nep.txt`` file needs to be provided. The model must be trained to predict the
    polarizability.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model in ``nep.txt`` format. Defaults to ``None``.
    debug
        Flag to toggle debug mode. Prints GPUMD output. Defaults to ``False``.

    Returns
    -------
    polarizability with shape ``(3, 3)``
    """
    if model_filename is None:
        raise ValueError('Model is undefined')

    model_type = _get_nep_contents(model_filename)[0]['model_type']
    if model_type != 'polarizability':
        raise ValueError(
            'A NEP model trained for predicting polarizability must be used.'
        )

    local_structure = structure.copy()
    natoms = len(local_structure)
    cell, symbols, positions, masses = _get_atomic_properties(local_structure)

    nepy = _setup_nepy(
        model_filename, natoms, cell, symbols, positions, masses, debug
    )
    # Components are in order xx yy zz xy yz zx.
    pol = nepy.get_polarizability()
    polarizability = np.array([
        [pol[0], pol[3], pol[5]],
        [pol[3], pol[1], pol[4]],
        [pol[5], pol[4], pol[2]]
    ])

    return polarizability


def get_dipole(
    structure: Atoms,
    model_filename: Optional[str] = None,
    debug: bool = False,
) -> np.ndarray:
    """Calculates the dipole for a given structure. A NEP model defined by a
    ``nep.txt`` file needs to be provided.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model in ``nep.txt`` format. Defaults to ``None``.
    debug
        Flag to toggle debug mode. Prints GPUMD output. Defaults to ``False``.

    Returns
    -------
    dipole with shape ``(3,)``
    """
    if model_filename is None:
        raise ValueError('Model is undefined')

    model_type = _get_nep_contents(model_filename)[0]['model_type']
    if model_type != 'dipole':
        raise ValueError('A NEP model trained for predicting dipoles must be used.')

    local_structure = structure.copy()
    natoms = len(local_structure)
    cell, symbols, positions, masses = _get_atomic_properties(local_structure)

    nepy = _setup_nepy(
        model_filename, natoms, cell, symbols, positions, masses, debug
    )
    dipole = np.array(nepy.get_dipole())

    return dipole


def get_dipole_gradient(
    structure: Atoms,
    model_filename: Optional[str] = None,
    backend: str = 'c++',
    method: str = 'central difference',
    displacement: float = 0.01,
    charge: float = 1.0,
    nep_command: str = 'nep',
    debug: bool = False,
) -> np.ndarray:
    """Calculates the dipole gradient for a given structure using finite differences.
    A NEP model defined by a `nep.txt` file needs to be provided.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model in ``nep.txt`` format. Defaults to ``None``.
    backend
        Backend to use for computing dipole gradient with finite differences.
        One of ``'c++'`` (CPU), ``'python'`` (CPU) and ``'nep'`` (GPU).
        Defaults to ``'c++'``.
    method
        Method for computing gradient with finite differences.
        One of 'forward difference' and 'central difference'.
        Defaults to 'central difference'
    displacement
        Displacement in Å to use for finite differences. Defaults to ``0.01``.
    charge
        System charge in units of the elemental charge.
        Used for correcting the dipoles before computing the gradient.
        Defaults to ``1.0``.
    nep_command
        Command for running the NEP executable. Defaults to ``'nep'``.
    debug
        Flag to toggle debug mode. Prints GPUMD output (if applicable). Defaults to ``False``.


    Returns
    -------
        dipole gradient with shape ``(N, 3, 3)``
    """
    if model_filename is None:
        raise ValueError('Model is undefined')

    model_type = _get_nep_contents(model_filename)[0]['model_type']
    if model_type != 'dipole':
        raise ValueError('A NEP model trained for predicting dipoles must be used.')

    local_structure = structure.copy()

    if backend == 'c++':
        dipole_gradient = _dipole_gradient_cpp(
            local_structure,
            model_filename,
            displacement=displacement,
            method=method,
            charge=charge,
            debug=debug,
        )
    elif backend == 'python':
        dipole_gradient = _dipole_gradient_python(
            local_structure,
            model_filename,
            displacement=displacement,
            charge=charge,
            method=method,
        )
    elif backend == 'nep':
        dipole_gradient = _dipole_gradient_nep(
            local_structure,
            model_filename,
            displacement=displacement,
            method=method,
            charge=charge,
            nep_command=nep_command,
        )
    else:
        raise ValueError(f'Invalid backend {backend}')
    return dipole_gradient


def _dipole_gradient_cpp(
    structure: Atoms,
    model_filename: str,
    method: str = 'central difference',
    displacement: float = 0.01,
    charge: float = 1.0,
    debug: bool = False,
) -> np.ndarray:
    """Calculates the dipole gradient with finite differences, using NEP_CPU.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model in ``nep.txt`` format.
    method
        Method for computing gradient with finite differences.
        One of ``'forward difference'`` and ``'central difference'``.
        Defaults to ``'central difference'``
    displacement
        Displacement in Å to use for finite differences. Defaults to ``0.01``.
    charge
        System charge in units of the elemental charge.
        Used for correcting the dipoles before computing the gradient.
        Defaults to ``1.0``.

    Returns
    -------
        dipole gradient with shape ``(N, 3, 3)``
    """
    if displacement <= 0:
        raise ValueError('Displacement must be > 0 Å')

    implemented_methods = {
        'forward difference': 0,
        'central difference': 1,
        'second order central difference': 2,
    }

    if method not in implemented_methods.keys():
        raise ValueError(f'Invalid method {method} for calculating gradient')

    local_structure = structure.copy()
    natoms = len(local_structure)
    cell, symbols, positions, masses = _get_atomic_properties(local_structure)
    nepy = _setup_nepy(
        model_filename, natoms, cell, symbols, positions, masses, debug
    )
    dipole_gradient = np.array(
        nepy.get_dipole_gradient(displacement, implemented_methods[method], charge)
    ).reshape(natoms, 3, 3)
    return dipole_gradient


def _dipole_gradient_python(
    structure: Atoms,
    model_filename: str,
    method: str = 'central difference',
    displacement: float = 0.01,
    charge: float = 1.0,
) -> np.ndarray:
    """Calculates the dipole gradient with finite differences, using the Python and get_dipole().

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model in ``nep.txt`` format.
    method
        Method for computing gradient with finite differences.
        One of ``'forward difference'`` and ``'central difference'``.
        Defaults to ``'central difference'``
    displacement
        Displacement in Å to use for finite differences. Defaults to ``0.01``.
    charge
        System charge in units of the elemental charge.
        Used for correcting the dipoles before computing the gradient.
        Defaults to ``1.0``.

    Returns
    -------
        dipole gradient with shape ``(N, 3, 3)``
    """
    if displacement <= 0:
        raise ValueError('Displacement must be > 0 Å')

    N = len(structure)
    if method == 'forward difference':
        # Correct all dipole by the permanent dipole, charge * center of mass
        d = (
            get_dipole(structure, model_filename)
            + charge * structure.get_center_of_mass()
        )
        d_forward = np.zeros((N, 3, 3))
        for atom in range(N):
            for cartesian in range(3):
                copy = structure.copy()
                positions = copy.get_positions()
                positions[atom, cartesian] += displacement
                copy.set_positions(positions)
                d_forward[atom, cartesian, :] = (
                    get_dipole(copy, model_filename)
                    + charge * copy.get_center_of_mass()
                )
        gradient = (d_forward - d[None, None, :]) / displacement

    elif method == 'central difference':
        d_forward = np.zeros((N, 3, 3))
        d_backward = np.zeros((N, 3, 3))
        for atom in range(N):
            for cartesian in range(3):
                # Forward displacements
                copy_forward = structure.copy()
                positions_forward = copy_forward.get_positions()
                positions_forward[atom, cartesian] += displacement
                copy_forward.set_positions(positions_forward)
                d_forward[atom, cartesian, :] = (
                    get_dipole(copy_forward, model_filename)
                    + charge * copy_forward.get_center_of_mass()
                )
                # Backwards displacement
                copy_backward = structure.copy()
                positions_backward = copy_backward.get_positions()
                positions_backward[atom, cartesian] -= displacement
                copy_backward.set_positions(positions_backward)
                d_backward[atom, cartesian, :] = (
                    get_dipole(copy_backward, model_filename)
                    + charge * copy_backward.get_center_of_mass()
                )
        gradient = (d_forward - d_backward) / (2 * displacement)
    elif method == 'second order central difference':
        # Coefficients from
        # https://en.wikipedia.org/wiki/Finite_difference_coefficient#Central_finite_difference
        d_forward_one_h = np.zeros((N, 3, 3))
        d_forward_two_h = np.zeros((N, 3, 3))
        d_backward_one_h = np.zeros((N, 3, 3))
        d_backward_two_h = np.zeros((N, 3, 3))
        for atom in range(N):
            for cartesian in range(3):
                copy = structure.copy()
                positions = copy.get_positions()
                # Forward displacements
                positions[atom, cartesian] += displacement  # + h
                copy.set_positions(positions)
                d_forward_one_h[atom, cartesian, :] = (
                    get_dipole(copy, model_filename)
                    + charge * copy.get_center_of_mass()
                )
                positions[atom, cartesian] += displacement  # + 2h total
                copy.set_positions(positions)
                d_forward_two_h[atom, cartesian, :] = (
                    get_dipole(copy, model_filename)
                    + charge * copy.get_center_of_mass()
                )
                # Backwards displacement
                positions[atom, cartesian] -= 3 * displacement  # 2h - 3h = -h
                copy.set_positions(positions)
                d_backward_one_h[atom, cartesian, :] = (
                    get_dipole(copy, model_filename)
                    + charge * copy.get_center_of_mass()
                )
                positions[atom, cartesian] -= displacement  # - 2h total
                copy.set_positions(positions)
                d_backward_two_h[atom, cartesian, :] = (
                    get_dipole(copy, model_filename)
                    + charge * copy.get_center_of_mass()
                )
            c0 = -1.0 / 12.0
            c1 = 2.0 / 3.0
            gradient = (
                c0 * d_forward_two_h
                + c1 * d_forward_one_h
                - c1 * d_backward_one_h
                - c0 * d_backward_two_h
            ) / displacement
    else:
        raise ValueError(f'Invalid method {method} for calculating gradient')
    return gradient


def _dipole_gradient_nep(
    structure: Atoms,
    model_filename: str,
    method: str = 'central difference',
    displacement: float = 0.01,
    charge: float = 1.0,
    nep_command: str = 'nep',
) -> np.ndarray:
    """Calculates the dipole gradient with finite differences, using the NEP executable.

    Parameters
    ----------
    structure
        Input structure
    model_filename
        Path to NEP model in ``nep.txt`` format.
    method
        Method for computing gradient with finite differences.
        One of ``'forward difference'`` and ``'central difference'``.
        Defaults to ``'central difference'``
    displacement
        Displacement in Å to use for finite differences. Defaults to 0.01 Å.
        Note that results are possibly unreliable for displacemen < 0.01,
        which might be due to rounding errors.
    charge
        System charge in units of the elemental charge.
        Used for correcting the dipoles before computing the gradient.
        Defaults to 1.0.
    nep_command
        Command for running the NEP executable. Defaults to ``'nep'``.


    Returns
    -------
        dipole gradient with shape ``(N, 3, 3)``
    """
    if displacement <= 0:
        raise ValueError('Displacement must be > 0 Å')

    if displacement < 1e-2:
        warnings.warn(
            'Dipole gradients with nep are unstable for displacements < 0.01 Å.'
        )

    N = len(structure)
    if method == 'forward difference':
        structure = _set_dummy_energy_forces(structure)
        structures = [structure]  # will hold 3N+1 structures
        # Correct for the constant dipole, by adding charge * center of mass
        corrections = np.zeros((3 * N + 1, 3))
        corrections[0] = charge * structure.get_center_of_mass()
        for atom in range(N):
            for cartesian in range(3):
                copy = structure.copy()
                positions = copy.get_positions()
                positions[atom, cartesian] += displacement
                copy.set_positions(positions)
                copy = _set_dummy_energy_forces(copy)
                structures.append(copy)
                corrections[1 + atom * 3 + cartesian] = (
                    charge * copy.get_center_of_mass()
                )

        dipoles = (
            _predict_dipole_batch(structures, model_filename, nep_command) * N
        )  # dipole/atom, shape (3N+1, 3)
        dipoles += corrections

        d = dipoles[0, :]
        d_forward = dipoles[1:].reshape(N, 3, 3)
        gradient = (d_forward - d[None, None, :]) / displacement

    elif method == 'central difference':
        structures_forward = []  # will hold 3N structures
        structures_backward = []  # will hold 3N structures

        # Correct for the constant dipole, by adding charge * center of mass
        corrections_forward = np.zeros((3 * N, 3))
        corrections_backward = np.zeros((3 * N, 3))

        for atom in range(N):
            for cartesian in range(3):
                # Forward displacements
                copy_forward = structure.copy()
                positions_forward = copy_forward.get_positions()
                positions_forward[atom, cartesian] += displacement
                copy_forward.set_positions(positions_forward)
                copy_forward = _set_dummy_energy_forces(copy_forward)
                structures_forward.append(copy_forward)
                corrections_forward[atom * 3 + cartesian] = (
                    charge * copy_forward.get_center_of_mass()
                )

                # Backwards displacement
                copy_backward = structure.copy()
                positions_backward = copy_backward.get_positions()
                positions_backward[atom, cartesian] -= displacement
                copy_backward.set_positions(positions_backward)
                copy_backward = _set_dummy_energy_forces(copy_backward)
                structures_backward.append(copy_backward)
                corrections_backward[atom * 3 + cartesian] = (
                    charge * copy_backward.get_center_of_mass()
                )

        structures = structures_forward + structures_backward
        dipoles = (
            _predict_dipole_batch(structures, model_filename, nep_command) * N
        )  # dipole/atom, shape (6N, 3)
        d_forward = dipoles[: 3 * N, :]
        d_backward = dipoles[3 * N :, :]

        d_forward += corrections_forward
        d_backward += corrections_backward

        d_forward = d_forward.reshape(N, 3, 3)
        d_backward = d_backward.reshape(N, 3, 3)

        gradient = (d_forward - d_backward) / (2 * displacement)
    else:
        raise ValueError(f'Invalid method {method} for calculating gradient')
    return gradient


def _set_dummy_energy_forces(structure: Atoms) -> Atoms:
    """Sets the energies and forces of structure to zero.

    Parameters
    ----------
    structure
        Input structure


    Returns
    ------- Copy of structure, with SinglePointCalculator with zero energy and force.
    """
    from ase.calculators.singlepoint import SinglePointCalculator

    copy = structure.copy()
    N = len(copy)
    energy = 0

    forces = np.zeros((N, 3))
    dummy = SinglePointCalculator(copy, **{'energy': energy, 'forces': forces})
    copy.calc = dummy
    return copy


def _predict_dipole_batch(
    structures: List[Atoms], model_filename: str, nep_command: str = 'nep'
) -> np.ndarray:
    """Predicts dipoles for a set of structures using the NEP executable
    Note that the units are in (dipole units)/atom.

    Parameters
    ----------
    structure
        Input structures
    model_filename
        Path to NEP model in ``nep.txt`` format.
    nep_command
        Command for running the NEP executable. Defaults to ``'nep'``.


    Returns
    ------- Predicted dipoles, with shape (len(structures), 3).
    """
    import shutil
    from os.path import join as join_path
    from subprocess import run
    from tempfile import TemporaryDirectory

    from calorine.nep import read_model, write_nepfile, write_structures

    with TemporaryDirectory() as directory:
        shutil.copy2(model_filename, join_path(directory, 'nep.txt'))
        model = read_model(model_filename)

        parameters = dict(
            prediction=1,
            mode=1,
            version=model.version,
            n_max=[model.n_max_radial, model.n_max_angular],
            cutoff=[model.radial_cutoff, model.angular_cutoff],
            basis_size=[model.n_basis_radial, model.n_basis_radial],
            l_max=[model.l_max_3b, model.l_max_4b, model.l_max_5b],
            neuron=model.n_neuron,
            type=[len(model.types), *model.types],
        )

        write_nepfile(parameters, directory)
        file = join_path(directory, 'train.xyz')
        write_structures(file, structures)

        # Execute nep
        completed = run([nep_command], cwd=directory, capture_output=True)
        completed.check_returncode()

        # Read results
        dipoles = np.loadtxt(join_path(directory, 'dipole_train.out'))
        if len(dipoles.shape) == 1:
            dipoles = dipoles.reshape(1, -1)
        return dipoles[:, :3]


def _check_components_polarizability_gradient(component: Union[str, List[str]]) -> List[int]:
    """
    Verifies that the selected components are ok.
    """
    allowed_components = {
            'x': 0,
            'y': 1,
            'z': 2,
    }

    # Check if chosen components are ok
    if component == 'full':
        components_to_compute = [0, 1, 2]
    else:
        components_list = [component] if isinstance(component, str) else component
        components_to_compute = []
        for c in components_list:
            if c in allowed_components.keys():
                components_to_compute.append(allowed_components[c])
            elif c == 'full':
                raise ValueError('Write ``component="full"`` to get all components.')
            else:
                raise ValueError(f'Invalid component {c}')
        assert len(components_list) == len(components_to_compute), \
            'Number of components to compute does not match.'
    return components_to_compute


def get_polarizability_gradient(
    structure: Atoms,
    model_filename: Optional[str] = None,
    displacement: float = 0.01,
    component: Union[str, List[str]] = 'full',
    debug: bool = False,
) -> np.ndarray:
    """Calculates the dipole gradient for a given structure using finite differences.
    A NEP model defined by a ``nep.txt`` file needs to be provided.
    This function computes the derivatives using the second-order central difference
    method with a C++ backend.

    Parameters
    ----------
    structure
        Input structure.
    model_filename
        Path to NEP model in ``nep.txt`` format. Defaults to ``None``.
    displacement
        Displacement in Å to use for finite differences. Defaults to ``0.01``.
    component
        Component or components of the polarizability tensor that the gradient
        should be computed for.
        The following components are available: ``x`, ``y``, ``z``, ``full``.
        Option ``full`` computes the derivative whilst moving the atoms in each Cartesian
        direction, which yields a tensor of shape ``(N, 3, 6)``.
        Multiple components may be specified.
        Defaults to ``full``.
    debug
        Flag to toggle debug mode. Prints GPUMD output (if applicable). Defaults to ``False``.


    Returns
    -------
        polarizability gradient with shape ``(N, C, 6)`` where ``C``
        is the number of components chosen.
    """
    if model_filename is None:
        raise ValueError('Model is undefined')

    model_type = _get_nep_contents(model_filename)[0]['model_type']
    if model_type != 'polarizability':
        raise ValueError('A NEP model trained for predicting polarizability must be used.')

    components_to_compute = _check_components_polarizability_gradient(component)

    local_structure = structure.copy()
    polarizability_gradient = _polarizability_gradient_cpp(
        local_structure,
        model_filename,
        displacement=displacement,
        components=components_to_compute,
        debug=debug,
    )
    return polarizability_gradient


def _polarizability_gradient_to_3x3(natoms, pg):
    """
    Converts a polarizability gradient tensor with
    shape (Natoms, 3, 6) to (Natoms, 3, 3, 3).
    The 6 items in the polarizability gradient have
    the following order from GPUMD: xx, yy, zz, xy, yz, zx.
    """
    gpumd_to_3x3_indices = np.array([
        [0, 3, 5],
        [3, 1, 4],
        [5, 4, 2]
    ])
    polgrad_3x3 = np.zeros((natoms, 3, 3, 3))
    for i in range(3):
        for j in range(3):
            index = gpumd_to_3x3_indices[i, j]
            polgrad_3x3[:, :, i, j] = pg[:, :, index]
    return polgrad_3x3


def _polarizability_gradient_cpp(
    structure: Atoms,
    model_filename: str,
    displacement: float,
    components: List[int],
    debug: bool = False,
) -> np.ndarray:
    """Calculates the polarizability gradient with finite differences, using NEP_CPU.

    Parameters
    ----------
    structure
        Input structure.
    model_filename
        Path to NEP model in ``nep.txt`` format.
    displacement
        Displacement in Å to use for finite differences. Defaults to ``0.01``.
    components
        List of components to compute. Integer values from 0 to 2 that corresponds
        to indices in the ordered list [x, y, z]. Index 1 corresponds to the
        derivative with regards to only the y positions of the atoms, and so forth.

    Returns
    -------
        dipole gradient with shape ``(N, len(components), 3, 3)``
    """
    if displacement <= 0:
        raise ValueError('Displacement must be > 0 Å')
    # TODO possibly use components later to only move atoms in one cartesian direction
    local_structure = structure.copy()
    natoms = len(local_structure)
    cell, symbols, positions, masses = _get_atomic_properties(local_structure)
    nepy = _setup_nepy(
        model_filename, natoms, cell, symbols, positions, masses, debug
    )
    pg = np.array(
        nepy.get_polarizability_gradient(displacement, components)
        ).reshape(natoms, 3, 6)
    # Convert to 3x3
    polarizability_gradient = _polarizability_gradient_to_3x3(natoms, pg)
    return polarizability_gradient[:, components, :, :]  # Only return the relevant components



import os
import shutil
import warnings
import tempfile
from collections.abc import Iterable
from typing import Any, List, Tuple, Union

import numpy as np
from ase import Atoms
from ase.calculators.calculator import FileIOCalculator, OldShellProfile
from ase.io import read as ase_read
from ase.units import GPa

from ..gpumd import write_xyz


class GPUMDShellProfile(OldShellProfile):
    """This class provides an ASE calculator for NEP calculations with
    GPUMD.

    Parameters
    ----------
    command : str
        Command to run GPUMD with.
        Default: ``gpumd``
    gpu_identifier_index : int, None
        Index that identifies the GPU that GPUNEP should be run with.
        Typically, NVIDIA GPUs are enumerated with integer indices.
        See https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#env-vars.
        Set to None in order to use all available GPUs. Note that GPUMD exit with an error
        when running with more than one GPU if your system is not large enough.
        Default: 1
    """
    def __init__(self, command : str, gpu_identifier_index: Union[int, None]):
        if gpu_identifier_index is not None:
            # Do not set a specific device to use = use all available GPUs
            self.cuda_environment_variables = f'CUDA_VISIBLE_DEVICES={gpu_identifier_index}'
            command_with_gpus = f'export {self.cuda_environment_variables} && ' + command
        else:
            command_with_gpus = command
        super().__init__(command_with_gpus)


class GPUNEP(FileIOCalculator):
    """This class provides an ASE calculator for NEP calculations with
    GPUMD.

    This calculator writes files that are input to the `gpumd`
    executable. It is thus likely to be slow if many calculations
    are to be performed.

    Parameters
    ----------
    model_filename : str
        Path to file in ``nep.txt`` format with model parameters.
    directory : str
        Directory to run GPUMD in. If None, a temporary directory
        will be created and removed once the calculations are finished.
        If specified, the directory will not be deleted. In the latter
        case, it is advisable to do no more than one calculation with
        this calculator (unless you know exactly what you are doing).
    label : str
        Label for this calculator.
    atoms : Atoms
        Atoms to attach to this calculator.
    command : str
        Command to run GPUMD with.
        Default: ``gpumd``
    gpu_identifier_index : int
        Index that identifies the GPU that GPUNEP should be run with.
        Typically, NVIDIA GPUs are enumerated with integer indices.
        See https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#env-vars.
        Set to None in order to use all available GPUs. Note that GPUMD exit with an error
        when running with more than one GPU if your system is not large enough.
        Default: 1


    Example
    -------

    >>> calc = GPUNEP('nep.txt')
    >>> atoms.calc = calc
    >>> atoms.get_potential_energy()
    """

    command = 'gpumd'
    implemented_properties = ['energy', 'forces', 'stress']
    discard_results_on_any_change = True

    # We use list of tuples to define parameters for
    # MD simulations. Looks like a dictionary, but sometimes
    # we want to repeat the same keyword.
    single_point_parameters = [('dump_thermo', 1),
                               ('dump_force', 1),
                               ('dump_position', 1),
                               ('velocity', 1e-24),
                               ('time_step', 1e-6),  # 1 zeptosecond
                               ('ensemble', 'nve'),
                               ('run', 1)]

    def __init__(self,
                 model_filename: str,
                 directory: str = None,
                 label: str = 'GPUNEP',
                 atoms: Atoms = None,
                 command: str = command,
                 gpu_identifier_index: Union[int, None] = 0
                 ):

        # Determine run command
        # Determine whether to save stdout or not
        if directory is None and '>' not in command:
            # No need to save stdout if we run in temporary directory
            command += ' > /dev/null'
        elif '>' not in command:
            command += ' > stdout'
        self.command = command
        self.model_filename = model_filename

        # Determine directory to run in
        self._use_temporary_directory = directory is None
        self._directory = directory
        if self._use_temporary_directory:
            self._make_new_tmp_directory()
        else:
            self._potential_path = os.path.relpath(
                os.path.abspath(self.model_filename), self._directory)

        # Override the profile in ~/.config/ase/config.ini.
        # See https://wiki.fysik.dtu.dk/ase/ase/
        #            calculators/calculators.html#calculator-configuration
        profile = GPUMDShellProfile(command, gpu_identifier_index)
        FileIOCalculator.__init__(self,
                                  directory=self._directory,
                                  label=label,
                                  atoms=atoms,
                                  profile=profile)

        # If the model file is missing we should abort immediately
        # such that we can provide a more clear error message
        # (otherwise the code would fail without telling what is wrong).
        if not os.path.exists(model_filename):
            raise FileNotFoundError(f'{model_filename} does not exist.')

        # Read species from nep.txt
        with open(model_filename, 'r') as f:
            for line in f:
                if 'nep' in line:
                    self.species = line.split()[2:]

    def run_custom_md(
        self,
        parameters: List[Tuple[str, Any]],
        return_last_atoms: bool = False,
        only_prepare: bool = False,
    ):
        """
        Run a custom MD simulation.

        Parameters
        ----------
        parameters
            Parameters to be specified in the run.in file.
            The potential keyword is set automatically, all other
            keywords need to be set via this argument.
            Example::

                    [('dump_thermo', 100),
                     ('dump_position', 1000),
                     ('velocity', 300),
                     ('time_step', 1),
                     ('ensemble', ['nvt_ber', 300, 300, 100]),
                     ('run', 10000)]

        return_last_atoms
            If ``True`` the last saved snapshot will be returned.
        only_prepare
            If ``True`` the necessary input files will be written
             but theMD run will not be executed.

        Returns
        -------
            The last snapshot if :attr:`return_last_atoms` is ``True``.
        """
        if self._use_temporary_directory:
            self._make_new_tmp_directory()

        if self._use_temporary_directory and not return_last_atoms:
            raise ValueError('Refusing to run in temporary directory '
                             'and not returning atoms; all results will be gone.')

        if self._use_temporary_directory and only_prepare:
            raise ValueError('Refusing to only prepare in temporary directory, '
                             'all files will be removed.')

        # Write files and run
        FileIOCalculator.write_input(self, self.atoms)
        self._write_runfile(parameters)
        write_xyz(filename=os.path.join(self._directory, 'model.xyz'),
                  structure=self.atoms)

        if only_prepare:
            return None

        # Execute the calculation.
        self.execute()

        # Extract last snapshot if needed
        if return_last_atoms:
            last_atoms = ase_read(os.path.join(self._directory, 'movie.xyz'),
                                  format='extxyz', index=-1)

        if self._use_temporary_directory:
            self._clean()

        if return_last_atoms:
            return last_atoms
        else:
            return None

    def write_input(self, atoms, properties=None, system_changes=None):
        """
        Write the input files necessary for a single-point calculation.
        """
        if self._use_temporary_directory:
            self._make_new_tmp_directory()
        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        self._write_runfile(parameters=self.single_point_parameters)
        write_xyz(filename=os.path.join(self._directory, 'model.xyz'),
                  structure=atoms)

    def _write_runfile(self, parameters):
        """Write run.in file to define input parameters for MD simulation.

        Parameters
        ----------
        parameters : dict
            Defines all key-value pairs used in run.in file
            (see GPUMD documentation for a complete list).
            Values can be either floats, integers, or lists/tuples.
        """
        if len(os.listdir(self._directory)) > 0:
            warnings.warn(f'{self._directory} is not empty.')

        with open(os.path.join(self._directory, 'run.in'), 'w') as f:
            # Custom potential is allowed but normally it can be deduced
            if 'potential' not in [keyval[0] for keyval in parameters]:
                f.write(f'potential {self._potential_path} \n')
            # Write all keywords with parameter(s)
            for key, val in parameters:
                f.write(f'{key} ')
                if isinstance(val, Iterable) and not isinstance(val, str):
                    for v in val:
                        f.write(f'{v} ')
                else:
                    f.write(f'{val}')
                f.write('\n')

    def get_potential_energy_and_stresses_from_file(self):
        """
        Extract potential energy (third column of last line in thermo.out) and stresses
        from thermo.out
        """
        data = np.loadtxt(os.path.join(self._directory, 'thermo.out'))
        if len(data.shape) == 1:
            line = data
        else:
            line = data[-1, :]

        # Energy
        energy = line[2]

        # Stress
        stress = [v for v in line[3:9]]
        stress = -GPa * np.array(stress)  # to eV/A^3

        if np.any(np.isnan(stress)) or np.isnan(energy):
            raise ValueError(f'Failed to extract energy and/or stresses:\n {line}')
        return energy, stress

    def _read_potential_energy_and_stresses(self):
        """Reads potential energy and stresses."""
        self.results['energy'], self.results['stress'] = \
            self.get_potential_energy_and_stresses_from_file()

    def get_forces_from_file(self):
        """
        Extract forces (in eV/A) from last snapshot in force.out
        """
        data = np.loadtxt(os.path.join(self._directory, 'force.out'))
        return data[-len(self.atoms):, :]

    def _read_forces(self):
        """Reads forces (the last snapshot in force.out) in eV/A"""
        self.results['forces'] = self.get_forces_from_file()

    def read_results(self):
        """
        Read results from last step of MD calculation.
        """
        self._read_potential_energy_and_stresses()
        self._read_forces()
        if self._use_temporary_directory:
            self._clean()

    def _clean(self):
        """
        Remove directory with calculations.
        """
        shutil.rmtree(self._directory)

    def _make_new_tmp_directory(self):
        """
        Create a new temporary directory.
        """
        # We do not need to create a new temporary directory
        # if the current one is empty
        if self._directory is None or \
           (os.path.isdir(self._directory) and len(os.listdir(self._directory)) > 0):
            self._directory = tempfile.mkdtemp()
        self._potential_path = os.path.relpath(os.path.abspath(self.model_filename),
                                               self._directory)

    def set_atoms(self, atoms):
        """
        Set Atoms object.
        Used also when attaching calculator to Atoms object.
        """
        self.atoms = atoms
        self.results = {}

    def set_directory(self, directory):
        """
        Set path to a new directory. This makes it possible to run
        several calculations with the same calculator while saving
        all results
        """
        self._directory = directory
        self._use_temporary_directory = False
        self._potential_path = os.path.relpath(os.path.abspath(self.model_filename),
                                               self._directory)



from .analysis import (analyze_data,
                       get_autocorrelation_function,
                       get_correlation_length,
                       get_error_estimate,
                       get_rtc_from_hac)
from .phonons import get_force_constants
from .structures import relax_structure
from .stiffness import get_elastic_stiffness_tensor

__all__ = ['analyze_data',
           'get_autocorrelation_function',
           'get_correlation_length',
           'get_error_estimate',
           'get_elastic_stiffness_tensor',
           'get_force_constants',
           'get_rtc_from_hac',
           'relax_structure']



# -*- coding: utf-8 -*-
from .io import (
    get_parity_data,
    read_loss,
    read_nepfile,
    read_structures,
    write_nepfile,
    write_structures,
)
from .model import read_model
from .nep import (
    get_descriptors,
    get_dipole,
    get_dipole_gradient,
    get_polarizability,
    get_polarizability_gradient,
    get_latent_space,
    get_potential_forces_and_virials,
)
from .training_factory import setup_training

__all__ = [
    'read_loss',
    'read_nepfile',
    'read_model',
    'read_structures',
    'get_parity_data',
    'get_descriptors',
    'get_dipole',
    'get_dipole_gradient',
    'get_polarizability',
    'get_polarizability_gradient',
    'get_latent_space',
    'get_potential_forces_and_virials',
    'setup_training',
    'write_nepfile',
    'write_structures',
]



from itertools import combinations_with_replacement, product
import numpy as np
from ase import Atoms
from ase.units import GPa
from .structures import relax_structure


def get_elastic_stiffness_tensor(structure: Atoms,
                                 clamped: bool = False,
                                 epsilon: float = 1e-3,
                                 **kwargs) -> np.ndarray:
    """Calculate and return the elastic stiffness tensor in units of GPa for the
    given structure in `Voigt form <https://en.wikipedia.org/wiki/Voigt_notation>`_.

    Parameters
    ----------
    structure
        input structure; should be fully relaxed
    clamped
        if ``False`` (default) return the *relaxed* elastic stiffness tensor;
        if ``True`` return the *clamped ion* elastic stiffness tensor
    epsilon
        magnitude of the applied strain
    kwargs
        keyword arguments forwarded to the :func:`relax_structure
        <calorine.tools.relax_structure>` function used for relaxing the
        structure when computing the relaxed stiffness tensor; it should not be
        necessary to change the default for the vast majority of use cases; use
        with care

    Returns
    -------
    Stiffness tensor in units of GPa
    """

    # set up of deformations
    deformations = []
    for i, j in combinations_with_replacement(range(9), r=2):
        for s1, s2 in product([-1, 1], repeat=2):
            S = np.zeros((3, 3))
            S.flat[i] = s1
            S.flat[j] = s2
            deformations.append(S)
    deformations = np.array(deformations)
    deformations *= epsilon

    # compute strain energies
    reference_energy = structure.get_potential_energy()
    energies = []
    for S in deformations:
        cell = structure.get_cell()
        cell += cell @ S.T
        deformed_structure = structure.copy()
        deformed_structure.calc = structure.calc
        deformed_structure.set_cell(cell, scale_atoms=True)
        if not clamped:
            relax_structure(deformed_structure, constant_cell=True, **kwargs)
        energy = deformed_structure.get_potential_energy()
        energies.append(energy - reference_energy)
    energies = np.array(energies)

    # extract stiffness tensor (full rank)
    SS = np.einsum('nij,nkl->nijkl', deformations, deformations)
    M = SS.reshape(len(SS), -1)
    M *= 0.5
    C, *_ = np.linalg.lstsq(M, energies, rcond=None)
    C = C.reshape(3, 3, 3, 3)
    C /= (structure.cell.volume * GPa)

    # convert stiffness tensor to Voigt form
    voigts = np.array([1, 1, 2, 2, 3, 3, 2, 3, 3, 1, 1, 2]).reshape(-1, 2) - 1
    Cv = np.zeros((6, 6))
    for i, j in product(range(6), repeat=2):
        v1 = voigts[i]
        v2 = voigts[j]
        Cv[i, j] = C[(*v1, *v2)]

    return Cv



from typing import Any, Dict

import numpy as np
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

try:
    from phonopy import Phonopy
    from phonopy.structure.atoms import PhonopyAtoms
except ModuleNotFoundError:  # pragma: no cover
    raise ModuleNotFoundError('phonopy (https://pypi.org/project/phonopy/) is '
                              'required in order to use the functionality '
                              'in the phonons module.'
                              )  # pragma: no cover


def get_force_constants(structure: Atoms,
                        calculator: SinglePointCalculator,
                        supercell_matrix: np.ndarray,
                        kwargs_phonopy: Dict[str, Any] = {},
                        kwargs_generate_displacements: Dict[str, Any] = {}) -> Phonopy:
    """
    Calculates the force constants for a given structure using
    `phonopy <https://phonopy.github.io/phonopy/>`_, which needs to be cited if this function
    is used for generating data for publication.
    The function returns a `Phonopy` object that can be used to calculate, e.g.,
    the phonon dispersion, the phonon density of states as well as related quantities such
    as the thermal displacements and the free energy.

    Parameters
    ----------
    structure
        structure for which to compute the phonon dispersion; usually this is a primitive cell
    calculator
        ASE calculator to use for the calculation of forces
    supercell_matrix
        specification of supercell size handed over to phonopy;
        should be a tuple of three values or a matrix
    kwargs_phonopy
        *Expert option*:
        keyword arguments used when initializing the `Phonopy` object;
        this includes, e.g., the tolerance used when determining the symmetry (`symprec`) and
        `parameters for the non-analytical corrections
        <https://phonopy.github.io/phonopy/phonopy-module.html#non-analytical-term-correction>`_
        (`nac_params`)
    kwargs_generate_displacements
        *Expert option*:
        keyword arguments to be handed over to the `generate_displacements` method;
        this includes in particular the `distance` keyword, which specifies the
        magnitude of the atomic displacement imposed when calculating the force constant matrix
    """

    # prepare primitive unit cell for phonopy
    structure_ph = PhonopyAtoms(symbols=structure.symbols,
                                cell=structure.cell,
                                scaled_positions=structure.get_scaled_positions(),
                                pbc=structure.pbc)

    # make sure we are using the masses intended by the user
    structure_ph.masses = structure.get_masses()

    # prepare supercells
    phonon = Phonopy(structure_ph, supercell_matrix, **kwargs_phonopy)
    phonon.generate_displacements(**kwargs_generate_displacements)

    # compute force constant matrix
    forces = []
    for structure_ph in phonon.supercells_with_displacements:
        structure_ase = Atoms(symbols=structure_ph.symbols,
                              cell=structure_ph.cell,
                              scaled_positions=structure_ph.scaled_positions,
                              pbc=structure.pbc)
        structure_ase.calc = calculator
        forces.append(structure_ase.get_forces().copy())

    phonon.forces = forces
    phonon.produce_force_constants()

    return phonon



from ase import Atoms
from ase.filters import FrechetCellFilter
from ase.optimize import BFGS, LBFGS, FIRE, GPMin
from ase.optimize.sciopt import SciPyFminBFGS
from ase.units import GPa


def relax_structure(structure: Atoms,
                    fmax: float = 0.001,
                    steps: int = 500,
                    minimizer: str = 'bfgs',
                    constant_cell: bool = False,
                    constant_volume: bool = False,
                    scalar_pressure: float = 0.0,
                    **kwargs) -> None:
    """Relaxes the given structure.

    Parameters
    ----------
    structure
        Atomic configuration to relax.
    fmax
        Stop relaxation if the absolute force for all atoms falls below this value.
    steps
        Maximum number of relaxation steps the minimizer is allowed to take.
    minimizer
        Minimizer to use; possible values: 'bfgs', 'lbfgs', 'fire', 'gpmin', 'bfgs-scipy'.
    constant_cell
        If True do not relax the cell or the volume.
    constant_volume
        If True relax the cell shape but keep the volume constant.
    kwargs
        Keyword arguments to be handed over to the minimizer; possible arguments can be found
        in the `ASE documentation <https://wiki.fysik.dtu.dk/ase/ase/optimize.html>`_
        https://wiki.fysik.dtu.dk/ase/ase/filters.html#the-frechetcellfilter-class.
    scalar_pressure
        External pressure in GPa.
    """
    if structure.calc is None:
        raise ValueError('Structure has no attached calculator object')
    if constant_cell:
        ucf = structure
    else:
        ucf = FrechetCellFilter(
            structure, constant_volume=constant_volume, scalar_pressure=scalar_pressure * GPa)
    kwargs['logfile'] = kwargs.get('logfile', None)
    if minimizer == 'bfgs':
        dyn = BFGS(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'lbfgs':
        dyn = LBFGS(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'bfgs-scipy':
        dyn = SciPyFminBFGS(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'fire':
        dyn = FIRE(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'gpmin':
        dyn = GPMin(ucf, **kwargs)
        dyn.run(fmax=fmax, steps=steps)
    else:
        raise ValueError(f'Unknown minimizer: {minimizer}')



from dataclasses import dataclass
from itertools import product
from typing import Dict, List, Tuple

import numpy as np


def _get_model_type(first_row: List[str]) -> str:
    """Parses a the first row of a ``nep.txt`` file,
    and returns the type of NEP model. Available types
    are ``potential``, ``dipole`` and ``polarizability``.

    Parameters
    ----------
    first_row
        first row of a NEP file, split by white space.
    """
    model_type = first_row[0]
    if 'dipole' in model_type:
        return 'dipole'
    elif 'polarizability' in model_type:
        return 'polarizability'
    return 'potential'


def _get_nep_contents(filename: str) -> Tuple[Dict, List]:
    """Parses a ``nep.txt`` file, and returns a dict describing the header
    and an unformatted list of all model parameters.
    Intended to be used by the :func:`read_model <calorine.nep.read_model>` function.

    Parameters
    ----------
    filename
        input file name
    """
    # parse file and split header and parameters
    header = []
    parameters = []
    nheader = 5  # 5 rows for NEP2, 6-7 rows for NEP3 onwards
    base_line = 3
    with open(filename) as f:
        for k, line in enumerate(f.readlines()):
            flds = line.split()
            assert len(flds) != 0, f'Empty line number {k}'
            if k == 0 and 'zbl' in flds[0]:
                base_line += 1
                nheader += 1
            if k == base_line and 'basis_size' in flds[0]:
                # Introduced in nep.txt after GPUMD v3.2
                nheader += 1
            if k < nheader:
                header.append(tuple(flds))
            elif len(flds) == 1:
                parameters.append(float(flds[0]))
            else:
                raise IOError(f'Failed to parse line {k} from {filename}')
    # compile data from the header into a dict
    data = {}
    for flds in header:
        if flds[0] in ['cutoff', 'zbl']:
            data[flds[0]] = tuple(map(float, flds[1:]))
        elif flds[0] in ['n_max', 'l_max', 'ANN', 'basis_size']:
            data[flds[0]] = tuple(map(int, flds[1:]))
        elif flds[0].startswith('nep'):
            version = flds[0].replace('nep', '').split('_')[0]
            version = int(version)
            data['version'] = version
            data['types'] = flds[2:]
            data['model_type'] = _get_model_type(flds)
        else:
            raise ValueError(f'Unknown field: {flds[0]}')
    return data, parameters


@dataclass
class Model:
    r"""Objects of this class represent a NEP model in a form suitable for
    inspection and manipulation. Typically a :class:`Model` object is instantiated
    by calling the :func:`read_model <calorine.nep.read_model>` function.

    Attributes
    ----------
    version : int
        NEP version.
    model_type: str
        One of ``potential``, ``dipole`` or ``polarizability``.
    types : Tuple[str, ...]
        Chemical species that this model represents.
    radial_cutoff : float
        The radial cutoff parameter in Å.
    angular_cutoff : float
        The angular cutoff parameter in Å.
    max_neighbors_radial : int
        Maximum number of neighbors in neighbor list for radial terms.
    max_neighbors_angular : int
        Maximum number of neighbors in neighbor list for angular terms.
    radial_typewise_cutoff_factor : float
        The radial cutoff factor if use_typewise_cutoff is used.
    angular_typewise_cutoff_factor : float
        The angular cutoff factor if use_typewise_cutoff is used.
    zbl : Tuple[float, float]
        Inner and outer cutoff for transition to ZBL potential.
    zbl_typewise_cutoff_factor : float
        Typewise cutoff when use_typewise_cutoff_zbl is used.
    n_basis_radial : int
        Number of radial basis functions :math:`n_\mathrm{basis}^\mathrm{R}`.
    n_basis_angular : int
        Number of angular basis functions :math:`n_\mathrm{basis}^\mathrm{A}`.
    n_max_radial : int
        Maximum order of Chebyshev polymonials included in
        radial expansion :math:`n_\mathrm{max}^\mathrm{R}`.
    n_max_angular : int
        Maximum order of Chebyshev polymonials included in
        angular expansion :math:`n_\mathrm{max}^\mathrm{A}`.
    l_max_3b : int
        Maximum expansion order for three-body terms :math:`l_\mathrm{max}^\mathrm{3b}`.
    l_max_4b : int
        Maximum expansion order for four-body terms :math:`l_\mathrm{max}^\mathrm{4b}`.
    l_max_5b : int
        Maximum expansion order for five-body terms :math:`l_\mathrm{max}^\mathrm{5b}`.
    n_descriptor_radial : int
        Dimension of radial part of descriptor.
    n_descriptor_angular : int
        Dimension of angular part of descriptor.
    n_neuron : int
        Number of neurons in hidden layer.
    n_parameters : int
        Total number of parameters including scalers (which are not fit parameters).
    n_descriptor_parameters : int
        Number of parameters in descriptor.
    n_ann_parameters : int
        Number of neural network weights.
    ann_parameters : Dict[Tuple[str, Dict[str, np.darray]]]
        Neural network weights.
    q_scaler : List[float]
        Scaling parameters.
    radial_descriptor_weights : Dict[Tuple[str, str], np.ndarray]
        Radial descriptor weights by combination of species; the array for each combination
        has dimensions of
        :math:`(n_\mathrm{max}^\mathrm{R}+1) \times (n_\mathrm{basis}^\mathrm{R}+1)`.
    angular_descriptor_weights : Dict[Tuple[str, str], np.ndarray]
        Angular descriptor weights by combination of species; the array for each combination
        has dimensions of
        :math:`(n_\mathrm{max}^\mathrm{A}+1) \times (n_\mathrm{basis}^\mathrm{A}+1)`.
    """

    version: int
    model_type: str
    types: Tuple[str, ...]

    radial_cutoff: float
    angular_cutoff: float

    n_basis_radial: int
    n_basis_angular: int
    n_max_radial: int
    n_max_angular: int
    l_max_3b: int
    l_max_4b: int
    l_max_5b: int
    n_descriptor_radial: int
    n_descriptor_angular: int

    n_neuron: int
    n_parameters: int
    n_descriptor_parameters: int
    n_ann_parameters: int
    ann_parameters: Dict[str, Dict[str, np.ndarray]]
    q_scaler: List[float]
    radial_descriptor_weights: Dict[Tuple[str, str], np.ndarray]
    angular_descriptor_weights: Dict[Tuple[str, str], np.ndarray]

    zbl: Tuple[float, float] = None
    zbl_typewise_cutoff_factor: float = None
    max_neighbors_radial: int = None
    max_neighbors_angular: int = None
    radial_typewise_cutoff_factor: float = None
    angular_typewise_cutoff_factor: float = None

    _special_fields = [
        'ann_parameters',
        'q_scaler',
        'radial_descriptor_weights',
        'angular_descriptor_weights',
    ]

    def __str__(self) -> str:
        s = []
        for fld in self.__dataclass_fields__:
            if fld not in self._special_fields:
                s += [f'{fld:22} : {getattr(self, fld)}']
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += ['<table border="1" class="dataframe"']
        s += [
            '<thead><tr><th style="text-align: left;">Field</th><th>Value</th></tr></thead>'
        ]
        s += ['<tbody>']
        for fld in self.__dataclass_fields__:
            if fld not in self._special_fields:
                s += [
                    f'<tr><td style="text-align: left;">{fld:22}</td>'
                    f'<td>{getattr(self, fld)}</td><tr>'
                ]
        for fld in self._special_fields:
            d = getattr(self, fld)
            if fld.endswith('descriptor_weights'):
                dim = list(d.values())[0].shape
            if fld == 'ann_parameters' and self.version == 4:
                dim = (len(self.types), len(list(d.values())[0]))
            else:
                dim = len(d)
            s += [
                f'<tr><td style="text-align: left;">Dimension of {fld:22}</td><td>{dim}</td><tr>'
            ]
        s += ['</tbody>']
        s += ['</table>']
        return ''.join(s)

    def write(self, filename: str) -> None:
        """Write NEP model to file in ``nep.txt`` format."""
        with open(filename, 'w') as f:
            # header
            version_name = f'nep{self.version}'
            if self.zbl is not None:
                version_name += '_zbl'
            elif self.model_type != 'potential':
                version_name += f'_{self.model_type}'
            f.write(f'{version_name} {len(self.types)} {" ".join(self.types)}\n')
            if self.zbl is not None:
                f.write(f'zbl {" ".join(map(str, self.zbl))}\n')
            f.write(f'cutoff {self.radial_cutoff} {self.angular_cutoff}')
            f.write(f' {self.max_neighbors_radial} {self.max_neighbors_angular}')
            if (
                self.radial_typewise_cutoff_factor is not None
                and self.angular_typewise_cutoff_factor is not None
            ):
                f.write(f' {self.radial_typewise_cutoff_factor}'
                        f' {self.angular_typewise_cutoff_factor}')
            if self.zbl_typewise_cutoff_factor:
                f.write(f' {self.zbl_typewise_cutoff_factor}')
            f.write('\n')
            f.write(f'n_max {self.n_max_radial} {self.n_max_angular}\n')
            f.write(f'basis_size {self.n_basis_radial} {self.n_basis_angular}\n')
            f.write(f'l_max {self.l_max_3b} {self.l_max_4b} {self.l_max_5b}\n')
            f.write(f'ANN {self.n_neuron} 0\n')

            # neural network weights
            keys = self.types if self.version in (4, 5) else ['all_species']
            suffixes = ['', '_polar'] if self.model_type == 'polarizability' else ['']
            for suffix in suffixes:
                for s in keys:
                    # Order: w0, b0, w1 (, b1 if NEP5)
                    # w0 indexed as: n*N_descriptor + nu
                    w0 = self.ann_parameters[s][f'w0{suffix}']
                    b0 = self.ann_parameters[s][f'b0{suffix}']
                    w1 = self.ann_parameters[s][f'w1{suffix}']
                    for n in range(self.n_neuron):
                        for nu in range(
                            self.n_descriptor_radial + self.n_descriptor_angular
                        ):
                            f.write(f'{w0[n, nu]:15.7e}\n')
                    for b in b0[:, 0]:
                        f.write(f'{b:15.7e}\n')
                    for v in w1[0, :]:
                        f.write(f'{v:15.7e}\n')
                    if self.version == 5:
                        b1 = self.ann_parameters[s][f'b1{suffix}']
                        f.write(f'{b1:15.7e}\n')
                b1 = self.ann_parameters[f'b1{suffix}']
                f.write(f'{b1:15.7e}\n')

            # descriptor weights
            mat = []
            for s1 in self.types:
                for s2 in self.types:
                    mat = np.hstack(
                        [mat, self.radial_descriptor_weights[(s1, s2)].flatten()]
                    )
                    mat = np.hstack(
                        [mat, self.angular_descriptor_weights[(s1, s2)].flatten()]
                    )
            n_types = len(self.types)
            n = int(len(mat) / (n_types * n_types))
            mat = mat.reshape((n_types * n_types, n)).T
            for v in mat.flatten():
                f.write(f'{v:15.7e}\n')

            # scaler
            for v in self.q_scaler:
                f.write(f'{v:15.7e}\n')


def read_model(filename: str) -> Model:
    """Parses a file in ``nep.txt`` format and returns the
    content in the form of a :class:`Model <calorine.nep.model.Model>`
    object.

    Parameters
    ----------
    filename
        Input file name.
    """
    data, parameters = _get_nep_contents(filename)

    # sanity checks
    for fld in ['cutoff', 'basis_size', 'n_max', 'l_max', 'ANN']:
        assert fld in data, f'Invalid model file; {fld} line is missing'
    assert data['version'] in [
        3,
        4,
        5,
    ], 'Invalid model file; only NEP versions 3, 4 and 5 are currently supported'

    # split up cutoff tuple
    assert len(data['cutoff']) in [4, 6, 7]
    data['radial_cutoff'] = data['cutoff'][0]
    data['angular_cutoff'] = data['cutoff'][1]
    data['max_neighbors_radial'] = int(data['cutoff'][2])
    data['max_neighbors_angular'] = int(data['cutoff'][3])
    if len(data['cutoff']) >= 6:
        data['radial_typewise_cutoff_factor'] = data['cutoff'][4]
        data['angular_typewise_cutoff_factor'] = data['cutoff'][5]
    if len(data['cutoff']) == 7:
        data['zbl_typewise_cutoff_factor'] = data['cutoff'][6]
    del data['cutoff']

    # split up basis_size tuple
    assert len(data['basis_size']) == 2
    data['n_basis_radial'] = data['basis_size'][0]
    data['n_basis_angular'] = data['basis_size'][1]
    del data['basis_size']

    # split up n_max tuple
    assert len(data['n_max']) == 2
    data['n_max_radial'] = data['n_max'][0]
    data['n_max_angular'] = data['n_max'][1]
    del data['n_max']

    # split up nl_max tuple
    len_l = len(data['l_max'])
    assert len_l in [1, 2, 3]
    data['l_max_3b'] = data['l_max'][0]
    data['l_max_4b'] = data['l_max'][1] if len_l > 1 else 0
    data['l_max_5b'] = data['l_max'][2] if len_l > 2 else 0
    del data['l_max']

    # compute dimensions of descriptor components
    data['n_descriptor_radial'] = data['n_max_radial'] + 1
    l_max_enh = data['l_max_3b'] + (data['l_max_4b'] > 0) + (data['l_max_5b'] > 0)
    data['n_descriptor_angular'] = (data['n_max_angular'] + 1) * l_max_enh
    n_descriptor = data['n_descriptor_radial'] + data['n_descriptor_angular']

    # compute number of parameters
    data['n_neuron'] = data['ANN'][0]
    del data['ANN']
    n_types = len(data['types'])
    if data['version'] == 3:
        n = 1
        n_bias = 1
    elif data['version'] == 4:
        # one hidden layer per atomic species
        n = n_types
        n_bias = 1
    else:  # NEP5
        # like nep4, but additionally has an
        # individual bias term in the output
        # layer for each species.
        n = n_types
        n_bias = 1 + n_types  # one global bias + one per species

    n_ann_input_weights = (n_descriptor + 1) * data['n_neuron']  # weights + bias
    n_ann_output_weights = data['n_neuron']  # only weights
    n_ann_parameters = (
        n_ann_input_weights + n_ann_output_weights
    ) * n + n_bias

    n_descriptor_weights = n_types**2 * (
        (data['n_max_radial'] + 1) * (data['n_basis_radial'] + 1)
        + (data['n_max_angular'] + 1) * (data['n_basis_angular'] + 1)
    )
    data['n_parameters'] = n_ann_parameters + n_descriptor_weights + n_descriptor
    is_polarizability_model = data['model_type'] == 'polarizability'
    if data['n_parameters'] + n_ann_parameters == len(parameters):
        data['n_parameters'] += n_ann_parameters
        assert is_polarizability_model, (
            'Model is not labelled as a polarizability model, but the number of '
            'parameters matches a polarizability model.\n'
            'If this is a polarizability model trained with GPUMD <=v3.8, please '
            'modify the header in the nep.txt file to read '
            f'`nep{data["version"]}_polarizability`.\n'
        )
    assert data['n_parameters'] == len(parameters), (
        'Parsing of parameters inconsistent; please submit a bug report\n'
        f'{data["n_parameters"]} != {len(parameters)}'
    )
    data['n_ann_parameters'] = n_ann_parameters

    # split up parameters into the ANN weights, descriptor weights, and scaling parameters
    n1 = n_ann_parameters
    n1 *= 2 if is_polarizability_model else 1
    n2 = n1 + n_descriptor_weights
    data['ann_parameters'] = parameters[:n1]
    descriptor_weights = np.array(parameters[n1:n2])
    data['q_scaler'] = parameters[n2:]

    # Group ANN parameters
    pars = {}
    n1 = 0
    n_network_params = n_ann_input_weights + n_ann_output_weights  # except last bias
    n_neuron = data['n_neuron']
    keys = data['types'] if data['version'] in (4, 5) else ['all_species']

    n_count = 2 if is_polarizability_model else 1
    for count in range(n_count):
        # if polarizability model, all parameters including bias are repeated
        # need to offset n1 by +1 to handle bias
        n1 += count
        for s in keys:
            # Get the parameters for the ANN; in the case of NEP4, there is effectively
            # one network per atomic species.
            ann_parameters = data['ann_parameters'][n1 : n1 + n_network_params]
            ann_input_weights = ann_parameters[:n_ann_input_weights]
            w0 = np.zeros((n_neuron, n_descriptor))
            w0[...] = np.nan
            b0 = np.zeros((n_neuron, 1))
            b0[...] = np.nan
            for n in range(n_neuron):
                for nu in range(n_descriptor):
                    w0[n, nu] = ann_input_weights[n * n_descriptor + nu]
            b0[:, 0] = ann_input_weights[n_neuron * n_descriptor :]

            assert np.all(
                w0.shape == (n_neuron, n_descriptor)
            ), f'w0 has invalid shape for key {s}; please submit a bug report'
            assert np.all(
                b0.shape == (n_neuron, 1)
            ), f'b0 has invalid shape for key {s}; please submit a bug report'
            assert not np.any(
                np.isnan(w0)
            ), f'some weights in w0 are nan for key {s}; please submit a bug report'
            assert not np.any(
                np.isnan(b0)
            ), f'some weights in b0 are nan for key {s}; please submit a bug report'

            ann_output_weights = ann_parameters[
                n_ann_input_weights : n_ann_input_weights + n_ann_output_weights
            ]

            w1 = np.zeros((1, n_neuron))
            w1[0, :] = ann_output_weights[:]
            assert np.all(
                w1.shape == (1, n_neuron)
            ), f'w1 has invalid shape for key {s}; please submit a bug report'
            assert not np.any(
                np.isnan(w1)
            ), f'some weights in w1 are nan for key {s}; please submit a bug report'

            if count == 0:
                pars[s] = dict(w0=w0, b0=b0, w1=w1)
            else:
                pars[s].update({'w0_polar': w0, 'b0_polar': b0, 'w1_polar': w1})
            # Jump to bias
            n1 += n_network_params
            if n_bias > 1:
                # For NEP5 models we additionally have one bias term per species.
                # Currently NEP5 only exists for potential models, but we'll
                # keep it here in case it gets added down the line.
                bias_label = 'b1' if count == 0 else 'b1_polar'
                pars[s][bias_label] = data['ann_parameters'][n1]
                n1 += 1
        # For NEP3 and NEP4 we only have one bias.
        # For NEP5 we have one bias per species, and one global.
        if count == 0:
            pars['b1'] = data['ann_parameters'][n1]
        else:
            pars['b1_polar'] = data['ann_parameters'][n1]
    sum = 0
    for s in pars.keys():
        if s.startswith('b1'):
            sum += 1
        else:
            sum += np.sum([np.count_nonzero(p) for p in pars[s].values()])
    assert sum == n_ann_parameters * n_count, (
        'Inconsistent number of parameters accounted for; please submit a bug report\n'
        f'{sum} != {n_ann_parameters}'
    )
    data['ann_parameters'] = pars

    # split up descriptor by chemical species and radial/angular
    data['n_descriptor_parameters'] = len(descriptor_weights)
    n = int(len(descriptor_weights) / (n_types * n_types))
    n_max_radial = data['n_max_radial']
    n_max_angular = data['n_max_angular']
    n_basis_radial = data['n_basis_radial']
    n_basis_angular = data['n_basis_angular']
    m = (n_max_radial + 1) * (n_basis_radial + 1)
    descriptor_weights = descriptor_weights.reshape((n, n_types * n_types)).T
    descriptor_weights_radial = descriptor_weights[:, :m]
    descriptor_weights_angular = descriptor_weights[:, m:]

    # add descriptors to data dict
    data['radial_descriptor_weights'] = {}
    data['angular_descriptor_weights'] = {}
    m = -1
    for i, j in product(range(n_types), repeat=2):
        m += 1
        s1, s2 = data['types'][i], data['types'][j]
        subdata = descriptor_weights_radial[m, :].reshape(
            (n_max_radial + 1, n_basis_radial + 1)
        )
        data['radial_descriptor_weights'][(s1, s2)] = subdata
        subdata = descriptor_weights_angular[m, :].reshape(
            (n_max_angular + 1, n_basis_angular + 1)
        )
        data['angular_descriptor_weights'][(s1, s2)] = subdata

    return Model(**data)



from .io import (
    read_hac, read_kappa, read_mcmd, read_runfile, read_thermo, read_thermodynamic_data,
    read_xyz, write_runfile, write_xyz,
)

__all__ = [
    'read_hac',
    'read_kappa',
    'read_mcmd',
    'read_runfile',
    'read_thermo',
    'read_thermodynamic_data',
    'read_xyz',
    'write_runfile',
    'write_xyz',
]



from os import makedirs
from os.path import exists, join as join_path
from typing import List, NamedTuple, Optional

import numpy as np
from ase import Atoms
from sklearn.model_selection import KFold

from .io import write_nepfile, write_structures


def setup_training(parameters: NamedTuple,
                   structures: List[Atoms],
                   enforced_structures: List[int] = [],
                   rootdir: str = '.',
                   mode: str = 'kfold',
                   n_splits: int = None,
                   train_fraction: float = None,
                   seed: int = 42,
                   overwrite: bool = False,
                   ) -> None:
    """Sets up the input files for training a NEP via the ``nep``
    executable of the GPUMD package.

    Parameters
    ----------
    parameters
        dictionary containing the parameters to be set in the nep.in file;
        see `here <https://gpumd.org/nep/input_parameters/index.html>`__
        for an overview of these parameters
    structures
        list of structures to be included
    enforced_structures
        structures that _must_ be included in the training set, provided in the form
        of a list of indices that refer to the content of the ``structures`` parameter
    rootdir
        root directory in which to create the input files
    mode
        how the test-train split is performed. Options: ``'kfold'`` and ``'bagging'``
    n_splits
        number of splits of the input structures in training and test sets that ought to be
        performed; by default no split will be done and all input structures will be used
        for training
    train_fraction
        fraction of structures to use for training when mode ``'bagging'`` is used
    seed
        random number generator seed to be used; this ensures reproducability
    overwrite
        if True overwrite the content of ``rootdir`` if it exists
    """
    if exists(rootdir) and not overwrite:
        raise FileExistsError('Output directory exists.'
                              ' Set overwrite=True in order to override this behavior.')

    if n_splits is not None and (n_splits <= 0 or n_splits > len(structures)):
        raise ValueError(f'n_splits ({n_splits}) must be positive and'
                         f' must not exceed {len(structures)}.')

    if mode == 'kfold' and train_fraction is not None:
        raise ValueError(f'train_fraction cannot be set when mode {mode} is used')
    elif mode == 'bagging' and (train_fraction <= 0 or train_fraction > 1):
        raise ValueError(f'train_fraction ({train_fraction}) must be in (0,1]')

    rs = np.random.RandomState(seed)
    _prepare_training(parameters, structures, enforced_structures,
                      rootdir, mode, n_splits, train_fraction, rs)


def _prepare_training(parameters: NamedTuple,
                      structures: List[Atoms],
                      enforced_structures: List[int],
                      rootdir: str,
                      mode: str,
                      n_splits: Optional[int],
                      train_fraction: Optional[float],
                      rs: np.random.RandomState) -> None:
    """Prepares training and test sets and writes structural data as well as parameters files.

    See class-level docstring for documentation of parameters.
    """
    dirname = join_path(rootdir, 'nepmodel_full')
    makedirs(dirname, exist_ok=True)
    _write_structures(structures, dirname, list(set(range(len(structures)))), [0])
    write_nepfile(parameters, dirname)

    if n_splits is None:
        return

    n_structures = len(structures)
    remaining_structures = list(set(range(n_structures)) - set(enforced_structures))

    if mode == 'kfold':
        kf = KFold(n_splits=n_splits, shuffle=True, random_state=rs)
        for k, (train_indices, test_indices) in enumerate(kf.split(remaining_structures)):
            # append enforced structures at the end of the training set
            train_selection = [remaining_structures[x] for x in list(train_indices)]
            test_selection = [remaining_structures[x] for x in list(test_indices)]

            # sanity check: make sure there is no overlap between train and test
            assert set(train_selection).intersection(set(test_selection)) == set(), \
                'Train and test set should not overlap'

            subdir = f'nepmodel_split{k+1}'
            dirname = join_path(rootdir, subdir)
            makedirs(dirname, exist_ok=True)
            _write_structures(structures, dirname, train_selection, test_selection)
            write_nepfile(parameters, dirname)

    elif mode == 'bagging':
        for k in range(n_splits):
            train_selection = rs.choice(
                remaining_structures,
                size=int(train_fraction * n_structures) - len(enforced_structures),
                replace=False)

            # append enforced structures at the end of the training set
            train_selection = list(train_selection)
            train_selection.extend(enforced_structures)

            # add the remaining structures to the test set
            test_selection = list(set(range(n_structures)) - set(train_selection))

            # sanity check: make sure there is no overlap between train and test
            assert set(train_selection).intersection(set(test_selection)) == set(), \
                'Train and test set should not overlap'

            dirname = join_path(rootdir, f'nepmodel_split{k+1}')
            makedirs(dirname, exist_ok=True)
            _write_structures(structures, dirname, train_selection, test_selection)
            write_nepfile(parameters, dirname)

    else:
        raise ValueError(f'Unknown value for mode: {mode}.')


def _write_structures(structures: List[Atoms],
                      dirname: str,
                      train_selection: List[int],
                      test_selection: List[int]):
    """Writes structures in format readable by nep executable.

    See class-level docstring for documentation of parameters.
    """
    write_structures(
        join_path(dirname, 'train.xyz'),
        [s for k, s in enumerate(structures) if k in train_selection])
    write_structures(
        join_path(dirname, 'test.xyz'),
        [s for k, s in enumerate(structures) if k in test_selection])
